"""
Utility code for gene sets handling
"""

import re
import csv
import hashlib

from .errors import AnnotationsError


GENESETS_TIDYCSV_HEADER = [
    "gene_set_name",
    "gene_set_description",
    "gene_symbol",
    "gene_description",
]


def read_gene_sets_tidycsv(gs_locator, context=None):
    """
    Read & parse the Tidy CSV format, applying validation checks for mandatory
    values, and de-duping rules.

    Format is a four-column CSV, with a mandatory header row, and optional "#" prefixed
    comments.  Format:

        gene_set_name, gene_set_description, gene_symbol, gene_description

    gene_set_name must be non-null; others are optional.

    Returns: a dictionary of the shape (values in angle-brackets vary):

        {
            <string, a gene set name>: {
                "geneset_name": <string, a gene set name>,
                "geneset_description": <a string or None>,
                "genes": [
                    {
                        "gene_symbol": <string, a gene symbol or name>,
                        "gene_description": <a string or None>
                    },
                    ...
                ]
            },
            ...
        }
    """

    class myDialect(csv.excel):
        skipinitialspace = False

    def just(n, seq):
        it = iter(seq)
        for _ in range(n - 1):
            yield next(it, "")
        yield tuple(it)

    messagefn = context["messagefn"] if context else (lambda x: None)

    gene_sets = {}
    with gs_locator.local_handle() as fname:
        with open(fname, newline="") as f:
            reader = csv.reader(f, dialect=myDialect())
            haveReadHeader = False
            lineno = 0
            for row in reader:
                lineno += 1
                # ignore empty rows
                if len(row) == 0:
                    continue
                # if row starts with '#' it is a comment
                if row[0].startswith("#"):
                    continue
                # if this is the first non-comment row, assume it is a header and validate
                # column names.  OK if the user has extra columns after our initial set.
                if not haveReadHeader:
                    if row[0:len(GENESETS_TIDYCSV_HEADER)] != GENESETS_TIDYCSV_HEADER:
                        raise AnnotationsError("Gene set CSV file missing the required column header.")
                    haveReadHeader = True
                    continue

                geneset_name, geneset_description, gene_symbol, gene_description, _ = just(5, row)
                if not geneset_name:
                    raise AnnotationsError(f"Gene set CSV missing required gene set name on line {lineno}")
                if (not gene_symbol) and gene_description:
                    messagefn(f"Warning: Missing gene name in gene set name {geneset_name} on line {lineno}.")

                if geneset_name in gene_sets:
                    gs = gene_sets[geneset_name]
                else:
                    gs = gene_sets[geneset_name] = {
                        "geneset_name": geneset_name,
                        "geneset_description": geneset_description,
                        "genes": [],
                    }
                # Use first geneset_description with a value
                if not gs["geneset_description"] and geneset_description:
                    gs["geneset_description"] = geneset_description
                # add the gene if the gene_symbol is defined
                if gene_symbol:
                    gs["genes"].append(
                        {
                            "gene_symbol": gene_symbol,
                            "gene_description": gene_description,
                        }
                    )

    return gene_sets


def write_gene_sets_tidycsv(f, genesets):
    """
    Convert the internal gene sets format (returned by read_gene_set_tidycsv) into
    the simple Tidy CSV.
    """
    writer = csv.writer(f, dialect="excel")
    writer.writerow(GENESETS_TIDYCSV_HEADER)
    for geneset in genesets:
        # genes may be empty, in which case we skip the gene set entirely
        genes = geneset["genes"]
        if not genes:
            writer.writerow([geneset["geneset_name"], geneset.get("geneset_description", ""), "", ""])
        else:
            writer.writerows(
                [
                    [
                        geneset["geneset_name"],
                        geneset.get("geneset_description", ""),
                        gene["gene_symbol"],
                        gene.get("gene_description", ""),
                    ]
                    for gene in genes
                ]
            )


def summarizeQueryHash(raw_query):
    """ generate a cache key (hash) from the raw query string """
    return hashlib.sha1(raw_query).hexdigest()


def validate_gene_sets(genesets, var_names, context=None):
    """
    Check validity of gene sets, return if correct, else raise error.
    May also modify the gene set for conditions that should be resolved,
    but which do not warrant a hard error.

    Argument gene sets may be either the REST OTA format (list of dicts) or the internal
    format (dict of dicts, keyed by the gene set name).

    Will return a modified gene sets (eg, remove warnings) of the same type as the
    provided argument. Ie, dict->dict, list->list

    Rules:

    0. All gene set names must be unique. [error]
    1. Gene set names must conform to the following: [error]
        * Names must be comprised of 1 or more ASCII characters 32-126
        * No leading or trailing spaces (ASCII 32)
        * No multi-space (ASCII 32) runs
    2. Gene symbols must be part of the current var_index. [warning]
       If gene symbol is not in the var_index, generate a warning and remove the symbol
       from the gene sets.
    3. Gene symbols must not be duplicated in a gene set.  [warning]
       Duplications will be silently de-duped.

    Items marked [error] will generate a hard error, causing the validation to fail.

    Items marked [warning] will generate a warning, and will be resolved without failing
    the validation (typically by removing the offending item from the gene sets).
    """

    messagefn = context["messagefn"] if context else (lambda x: None)

    # accept genesets args as either the internal (dict) or REST (list) format,
    # as they are identical except for the dict being keyed by geneset_name.
    if not isinstance(genesets, dict) and not isinstance(genesets, list):
        raise ValueError("Gene sets must be either dict or list.")
    genesets_iterable = genesets if isinstance(genesets, list) else genesets.values()

    # 0. check for uniqueness of geneset names
    geneset_names = [gs["geneset_name"] for gs in genesets_iterable]
    if len(set(geneset_names)) != len(geneset_names):
        raise KeyError("All gene set names must be unique.")

    # 1. check gene set character set and format
    illegal_name = re.compile(r"^\s|  |[\u0000-\u001F\u007F-\uFFFF]|\s$")
    for name in geneset_names:
        if type(name) != str or len(name) == 0:
            raise KeyError("Gene set names must be non-null string.")
        if illegal_name.search(name):
            messagefn(
                "Error: "
                f"Gene set name {name} "
                "is not valid. Leading, trailing, and multiple spaces within a name are not allowed."
            )
            raise KeyError(
                "Gene set name is not valid. Leading, trailing, and multiple spaces within a name are not allowed."
            )

    # 2. & 3. check for duplicate gene symbols, and those not present in the dataset. They will
    # generate a warning and be removed.
    for geneset in genesets_iterable:
        if not isinstance(geneset, dict):
            raise ValueError("Each gene set must be a dict.")
        geneset_name = geneset["geneset_name"]
        genes = geneset["genes"]
        if not isinstance(genes, list):
            raise ValueError("Gene set genes field must be a list")
        geneset.setdefault("geneset_description", "")
        gene_symbol_already_seen = set()
        new_genes = []
        for gene in genes:
            gene_symbol = gene["gene_symbol"]
            if not isinstance(gene_symbol, str) or len(gene_symbol) == 0:
                raise ValueError("Gene symbol must be non-null string.")
            if gene_symbol in gene_symbol_already_seen:
                # duplicate check
                messagefn(
                    f"Warning: a duplicate of gene {gene_symbol} was found in gene set {geneset_name}, "
                    "and will be ignored."
                )
                continue

            if gene_symbol not in var_names:
                messagefn(
                    f"Warning: {gene_symbol}, used in gene set {geneset_name}, "
                    "was not found in the dataset and will be ignored."
                )
                continue

            gene_symbol_already_seen.add(gene_symbol)
            gene.setdefault("gene_description", "")
            new_genes.append(gene)

        geneset["genes"] = new_genes

    return genesets



# import the built in auth types so they can be registered

import backend.server.auth.auth_none  # noqa: F401
import backend.server.auth.auth_test  # noqa: F401
import backend.server.auth.auth_session  # noqa: F401



"""
Corpora schema conventions support.  Helper functions for reading.

https://github.com/chanzuckerberg/corpora-data-portal/blob/main/backend/schema/corpora_schema.md

https://github.com/chanzuckerberg/corpora-data-portal/blob/main/backend/schema/corpora_schema_h5ad_implementation.md
"""
import collections
import json

from backend.server.cli.upgrade import validate_version_str
from backend.server.common.utils.corpora_constants import CorporaConstants


def corpora_get_versions_from_anndata(adata):
    """
    Given an AnnData object, return:
        * None - if not a Corpora object
        * [ corpora_schema_version, corpora_encoding_version ] - if a Corpora object

    Implements the identification protocol defined in the specification.
    """

    # per Corpora AnnData spec, this is a corpora file if the following is true
    if "version" not in adata.uns_keys():
        return None
    version = adata.uns["version"]
    if not isinstance(version, collections.abc.Mapping) or "corpora_schema_version" not in version:
        return None

    corpora_schema_version = version.get("corpora_schema_version")
    corpora_encoding_version = version.get("corpora_encoding_version")

    # TODO: spec says these must be SEMVER values, so check.
    if validate_version_str(corpora_schema_version) and validate_version_str(corpora_encoding_version):
        return [corpora_schema_version, corpora_encoding_version]


def corpora_is_version_supported(corpora_schema_version, corpora_encoding_version):
    return (
        corpora_schema_version
        and corpora_encoding_version
        and corpora_schema_version.startswith("1.")
        and corpora_encoding_version.startswith("0.1.")
    )


def corpora_get_props_from_anndata(adata):
    """
    Get Corpora dataset properties from an AnnData
    """
    versions = corpora_get_versions_from_anndata(adata)
    if versions is None:
        return None
    [corpora_schema_version, corpora_encoding_version] = versions
    version_is_supported = corpora_is_version_supported(corpora_schema_version, corpora_encoding_version)
    if not version_is_supported:
        raise ValueError("Unsupported Corpora schema version")

    corpora_props = {}
    for key in CorporaConstants.REQUIRED_SIMPLE_METADATA_FIELDS:
        if key not in adata.uns:
            raise KeyError(f"missing Corpora schema field {key}")
        corpora_props[key] = adata.uns[key]

    for key in CorporaConstants.OPTIONAL_JSON_ENCODED_METADATA_FIELD:
        if key not in adata.uns:
            continue
        try:
            corpora_props[key] = json.loads(adata.uns[key])
        except json.JSONDecodeError:
            raise json.JSONDecodeError(f"Corpora schema field {key} is expected to be a valid JSON string")

    for key in CorporaConstants.OPTIONAL_SIMPLE_METADATA_FIELDS:
        if key in adata.uns:
            corpora_props[key] = adata.uns[key]

    return corpora_props



import os
import tempfile
import fsspec
from datetime import datetime
import boto3
import botocore
from urllib.parse import urlparse


class DataLocator:
    """
    DataLocator is a simple wrapper around fsspec functionality, and provides a
    set of functions to encapsulate a data location (URI or path), interogate
    metadata about the object at that location (size, existance, etc) and
    access the underlying data.

    https://filesystem-spec.readthedocs.io/en/latest/index.html

    Example:
        dl = DataLocator("/tmp/foo.h5ad")
        if dl.exists():
            print(dl.size())
            with dl.open() as f:
                thecontents = f.read()

    DataLocator will accept a URI or native path.  Error handling is as defined
    in fsspec.

    """

    def __init__(self, uri_or_path, region_name=None):
        if isinstance(uri_or_path, DataLocator):
            locator = uri_or_path
            self.uri_or_path = locator.uri_or_path
            self.protocol = locator.protocol
            self.path = locator.path
            self.cname = locator.cname
        else:
            self.uri_or_path = uri_or_path
            self.protocol, self.path = DataLocator._get_protocol_and_path(uri_or_path)
            # work-around for LocalFileSystem not treating file: and None as the same scheme/protocol
            self.cname = self.path if self.protocol == "file" else self.uri_or_path

        # fsspec.filesystem will throw RuntimeError if the protocol is unsupported
        if self.protocol == "s3":
            if region_name:
                config_kwargs = dict(region_name=region_name)
                self.fs = fsspec.filesystem(self.protocol, listings_expiry_time=30, config_kwargs=config_kwargs)
            else:
                self.fs = fsspec.filesystem(self.protocol, listings_expiry_time=30)
        else:
            self.fs = fsspec.filesystem(self.protocol)

    def __repr__(self):
        return f"DataLocator(protocol={self.protocol}, cname={self.cname}, "
        f"path={self.path}, uri_or_path={self.uri_or_path})"

    @staticmethod
    def _get_protocol_and_path(uri_or_path):
        if "://" in uri_or_path:
            protocol, path = uri_or_path.split("://", 1)
            # windows!!!  Ignore single letter drive identifiers,
            # eg, G:\foo.txt
            if len(protocol) > 1:
                return protocol, path
        return None, uri_or_path

    def exists(self):
        return self.fs.exists(self.cname)

    def size(self):
        return self.fs.size(self.cname)

    def lastmodtime(self):
        """ return datetime object representing last modification time, or None if unavailable """
        info = self.fs.info(self.cname)
        if self.islocal() and info is not None:
            return datetime.fromtimestamp(info["mtime"])
        else:
            return getattr(info, "LastModified", None)

    def abspath(self):
        """
        return the absolute path for the locator - only really does something
        for file: protocol, as all others are already absolute
        """
        if self.islocal():
            return os.path.abspath(self.path)
        else:
            return self.uri_or_path

    def isfile(self):
        return self.fs.isfile(self.cname)

    def open(self, *args):
        return self.fs.open(self.uri_or_path, *args)

    def islocal(self):
        return self.protocol is None or self.protocol == "file"

    def local_handle(self):
        if self.islocal():
            return LocalFilePath(self.path)

        # if not local, create a tmp file system object to contain the data,
        # and clean it up when done.  If the path has a suffix/extension,
        # do our best to create a file with the same.
        ext = os.path.splitext(self.path)
        suffix = None if ext[1] == "" else ext[1]
        with self.open() as src, tempfile.NamedTemporaryFile(prefix="cellxgene_", suffix=suffix, delete=False) as tmp:
            tmp.write(src.read())
            tmp.close()
            src.close()
            tmp_path = tmp.name
            return LocalFilePath(tmp_path, delete=True)

    def ls(self):
        paths = self.fs.ls(self.uri_or_path)
        return [os.path.basename(p) for p in paths]


class LocalFilePath:
    def __init__(self, tmp_path, delete=False):
        self.tmp_path = tmp_path
        self.delete = delete

    def __enter__(self):
        return self.tmp_path

    def __exit__(self, *args):
        if self.delete:
            os.unlink(self.tmp_path)


def discover_s3_region_name(uri):
    """If this is an s3 protocol, discover and return the (aws) region name.
    If a return name could not be discovered, or if the uri is not an s3 protocol, return None."""

    protocol, _ = DataLocator._get_protocol_and_path(uri)
    if protocol == "s3":
        bucket = urlparse(uri).netloc
        client = boto3.client("s3")
        try:
            res = client.head_bucket(Bucket=bucket)
        except botocore.exceptions.ClientError:
            return None

        region = res.get("ResponseMetadata", {}).get("HTTPHeaders", {}).get("x-amz-bucket-region")
        if region:
            return region
        else:
            return None

    return None



import base64
import os
import re
import threading
from datetime import datetime
from hashlib import blake2b
import json

import pandas as pd
from flask import session, has_request_context, current_app, request

from backend.server import __version__ as cellxgene_version
from backend.server.common.annotations.annotations import Annotations
from backend.common.genesets import read_gene_sets_tidycsv
from backend.common.errors import AnnotationsError, ObsoleteRequest
from backend.common.utils.data_locator import DataLocator


class AnnotationsLocalFile(Annotations):
    CXG_ANNO_COLLECTION = "cxg_anno_collection"

    def __init__(self, config, output_dir, label_output_file, gene_sets_output_file, reembed_parameters_output_file):
        super().__init__(config)
        self.output_dir = output_dir
        self.label_output_file = label_output_file
        self.gene_sets_output_file = gene_sets_output_file
        self.reembed_parameters_output_file = reembed_parameters_output_file
        # lock used to protect label file write ops
        self.label_lock = threading.RLock()
        self.gene_sets_lock = threading.RLock()
        self.reembed_parameters_lock = threading.RLock()

        # cache the most recent cell labels/annotations.
        self.last_label_fname = None
        self.last_labels = None

        # cache the most recent gene sets.
        self.last_geneset_fname = None
        self.last_geneset = None

        # txn ID - used to de-dup geneset writes
        self.last_geneset_tid = 0

    def is_safe_collection_name(self, name):
        """
        return true if this is a safe collection name
        this is ultra conservative. If we want to allow full legal file name syntax,
        we could look at modules like `pathvalidate`
        """
        if name is None:
            return False
        return re.match(r"^[\w\-]+$", name) is not None

    def set_collection(self, name):
        session[self.CXG_ANNO_COLLECTION] = name
        session.permanent = True

    def get_collection(self):
        if session is None:
            return None
        return session.get(self.CXG_ANNO_COLLECTION)

    def read_labels(self, data_adaptor):
        self.check_user_annotations_enabled()  # raises

        if has_request_context():
            if not current_app.auth.is_user_authenticated():
                return pd.DataFrame()

        fname = self._get_celllabels_filename(data_adaptor)
        with self.label_lock:
            if fname is not None and os.path.exists(fname) and os.path.getsize(fname) > 0:
                # returned the cached labels if possible, otherwise read them from the file
                if fname == self.last_label_fname:
                    return self.last_labels
                else:
                    labels = pd.read_csv(
                        fname, dtype="category", index_col=0, header=0, comment="#", keep_default_na=False
                    )
                    # update the cache
                    self.last_label_fname = fname
                    self.last_labels = labels
                    return labels
            else:
                return pd.DataFrame()

    def write_labels(self, df, data_adaptor):
        self.check_user_annotations_enabled()  # raises

        # update our internal state and save it.  Multi-threading often enabled,
        # so treat this as a critical section.
        with self.label_lock:
            lastmod = data_adaptor.get_last_mod_time()
            lastmodstr = "'unknown'" if lastmod is None else lastmod.isoformat(timespec="seconds")
            header = (
                f"# Annotations generated on {datetime.now().isoformat(timespec='seconds')} "
                f"using cellxgene version {cellxgene_version}\n"
                f"# Input data file was {data_adaptor.get_location()}, "
                f"which was last modified on {lastmodstr}\n"
            )

            fname = self._get_celllabels_filename(data_adaptor)
            self._backup(fname)
            if not df.empty:
                with open(fname, "w", newline="") as f:
                    if header is not None:
                        f.write(header)
                    df.to_csv(f)
            else:
                open(fname, "w").close()

            # update the cache
            self.last_label_fname = fname
            self.last_labels = df

    def read_reembed_parameters(self,data_adaptor,context=None):
        if has_request_context():
            if not current_app.auth.is_user_authenticated():
                return ({})
        
        fname = self._get_reembed_parameters_filename(data_adaptor)
        with self.reembed_parameters_lock:
            try:
                with open(fname) as json_file:
                    parameters = json.load(json_file)
            except FileNotFoundError as error:
                parameters = {}
        return parameters
    
    def read_gene_sets(self, data_adaptor, context=None):
        if has_request_context():
            if not current_app.auth.is_user_authenticated():
                return ({}, self.last_geneset_tid)

        fname = self._get_genesets_filename(data_adaptor)
        gene_sets = {}
        tid = None
        with self.gene_sets_lock:
            tid = self.last_geneset_tid  # inside the critical section
            if fname is not None and os.path.exists(fname) and os.path.getsize(fname) > 0:
                # return the cached genesets if possible, otherwise read from file and validate them
                if fname == self.last_geneset_fname:
                    gene_sets = self.last_geneset
                else:
                    # read
                    gene_sets = read_gene_sets_tidycsv(DataLocator(fname), context)

                    # validate
                    gene_sets = data_adaptor.check_new_gene_sets(gene_sets, context)

                    # update cache
                    self.last_geneset_fname = fname
                    self.last_geneset = gene_sets

        return (gene_sets, tid)
    
    def write_reembed_parameters(self,parameters,data_adaptor):
        with self.reembed_parameters_lock:
            fname = self._get_reembed_parameters_filename(data_adaptor)
            with open(fname, 'w') as outfile:
                json.dump(parameters, outfile)

    def write_gene_sets(self, gene_sets, tid, data_adaptor):
        self.check_gene_sets_save_enabled()  # raises

        if type(tid) != int or tid < 0:
            raise ValueError("tid must be a positive integer")

        # may raise
        gene_sets = data_adaptor.check_new_gene_sets(gene_sets)

        with self.gene_sets_lock:
            # skip if the request is stale
            if tid is not None:
                if tid <= self.last_geneset_tid:
                    raise ObsoleteRequest("TID is stale.")
                self.last_geneset_tid = tid

            lastmod = data_adaptor.get_last_mod_time()
            lastmodstr = "'unknown'" if lastmod is None else lastmod.isoformat(timespec="seconds")
            header = (
                f"# Gene set generated on {datetime.now().isoformat(timespec='seconds')} "
                f"using cellxgene version {cellxgene_version}\n"
                f"# Input data file was {data_adaptor.get_location()}, "
                f"which was last modified on {lastmodstr}\n"
            )

            fname = self._get_genesets_filename(data_adaptor)
            self._backup(fname)
            with open(fname, "w", newline="") as f:
                f.write(header)
                f.write(self.gene_sets_to_csv(gene_sets))

            # update the cache
            self.last_geneset_fname = fname
            self.last_geneset = gene_sets if type(gene_sets) == dict else {g["geneset_name"]: g for g in gene_sets}

    def _get_userdata_idhash(self, data_adaptor):
        """
        Return a short hash that weakly identifies the user and dataset.
        Used to create safe annotations output file names.
        """
        useGuest = False
        if 'profile' not in session:
            useGuest = True
        elif 'excxgLoggedIn' not in session['profile']:
            useGuest = True

        if useGuest:
            id = (data_adaptor.get_location()).encode()
            guest_idhash = base64.b32encode(blake2b(id, digest_size=5).digest()).decode("utf-8")            
            idhash = guest_idhash
        else:
            uid = session['profile']['sub']
            id = (uid + data_adaptor.get_location()).encode()
            idhash = base64.b32encode(blake2b(id, digest_size=5).digest()).decode("utf-8")
        return idhash

    def _get_output_dir(self):
        if self.output_dir:
            return self.output_dir

        output_file = (self.label_output_file or self.gene_sets_output_file) or self.reembed_parameters_output_file
        if output_file:
            return os.path.dirname(os.path.abspath(output_file))

        return os.getcwd()

    def _get_celllabels_filename(self, data_adaptor):
        """ return the current annotation file name """
        if self.label_output_file:
            return self.label_output_file

        return self._get_filename(data_adaptor, "cell-labels")+".csv"

    def _get_genesets_filename(self, data_adaptor):
        """ return the current gene sets file name """
        if self.gene_sets_output_file:
            return self.gene_sets_output_file

        return self._get_filename(data_adaptor, "gene-sets")+".csv"

    def _get_reembed_parameters_filename(self, data_adaptor):
        """ return the current reembed parameters file name """
        if self.reembed_parameters_output_file:
            return self.reembed_parameters_output_file
        return self._get_filename(data_adaptor, "reembed-parameters")+".json"

    def _get_filename(self, data_adaptor, anno_name):
        # we need to generate a file name, which we can only do if we have a UID and collection name
        if session is None:
            raise AnnotationsError("unable to determine file name for annotations")

        if data_adaptor is None:
            raise AnnotationsError("unable to determine file name for annotations")

        idhash = self._get_userdata_idhash(data_adaptor)
        return os.path.join(self._get_output_dir(), f"{idhash}/{anno_name}")

    def _backup(self, fname, max_backups=9):
        """
        save N backups of file to backup_dir.
            1. fname -> backup_dir/fname-TIME
            2. delete excess files in backup_dir
        """
        root, ext = os.path.splitext(fname)
        backup_dir = f"{root}-backups"

        # Make sure there is work to do
        if not os.path.exists(fname):
            return

        # Ensure backup_dir exists
        if not os.path.exists(backup_dir):
            os.mkdir(backup_dir)

        # Save current file to backup_dir
        fname_base = os.path.basename(fname)
        fname_base_root, fname_base_ext = os.path.splitext(fname_base)
        # don't use ISO standard time format, as it contains characters illegal on some filesytems.
        nowish = datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
        backup_fname = os.path.join(backup_dir, f"{fname_base_root}-{nowish}{fname_base_ext}")
        if os.path.exists(backup_fname):
            os.remove(backup_fname)
        os.rename(fname, backup_fname)

        # prune the backup_dir to max number of backup files, keeping the most recent backups
        backups = list(filter(lambda s: s.startswith(fname_base_root), os.listdir(backup_dir)))
        excess_count = len(backups) - max_backups
        if excess_count > 0:
            backups.sort()
            for bu in backups[0:excess_count]:
                os.remove(os.path.join(backup_dir, bu))

    def update_parameters(self, parameters, data_adaptor):
        params = {}
        params["annotations"] = self.user_annotations_enabled()
        params["annotations_genesets_readonly"] = not self.gene_sets_save_enabled()
        params["annotations_genesets_name_is_read_only"] = self.gene_sets_output_file is not None
        params["user_annotation_collection_name_enabled"] = True

        if self.ontology_data:
            params["annotations_cell_ontology_enabled"] = True
            params["annotations_cell_ontology_terms"] = self.ontology_data
        else:
            params["annotations_cell_ontology_enabled"] = False

        if self.label_output_file is not None:
            # user has hard-wired the name of the annotation cell label data collection
            fname = os.path.basename(self.label_output_file)
            collection_fname = os.path.splitext(fname)[0]
            params["annotations-data-collection-is-read-only"] = True
            params["annotations-data-collection-name"] = collection_fname

        elif session is not None:
            collection = self.get_collection()
            params["annotations-data-collection-is-read-only"] = not self.user_annotations_enabled()
            params["annotations-data-collection-name"] = collection

        if current_app.auth.is_user_authenticated():
            params["annotations-user-data-idhash"] = self._get_userdata_idhash(data_adaptor)

        parameters.update(params)



class CorporaConstants(object):
    REQUIRED_SIMPLE_METADATA_FIELDS = [
        "version",
        "title",
        "layer_descriptions",
        "organism",
        "organism_ontology_term_id",
    ]

    # The Corpora specification requires some values encoded as JSON due to the inability of AnnData to store complex
    # types.
    OPTIONAL_JSON_ENCODED_METADATA_FIELD = ["contributors", "project_links"]

    OPTIONAL_SIMPLE_METADATA_FIELDS = [
        "preprint_doi",
        "publication_doi",
        "default_embedding",
        "default_field",
        "tags",
        "project_name",
        "project_description",
    ]



from backend.server.auth.auth import AuthTypeBase, AuthTypeFactory


class AuthTypeNone(AuthTypeBase):
    def __init__(self, app_config):
        super().__init__()

    def is_valid_authentication_type(self):
        return False

    def complete_setup(self, app):
        pass

    def is_user_authenticated(self):
        return True

    def get_user_id(self):
        return None

    def get_user_name(self):
        return None

    def get_user_email(self):
        return None


AuthTypeFactory.register(None, AuthTypeNone)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: NetEncoding

import flatbuffers

class Uint32Array(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsUint32Array(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Uint32Array()
        x.Init(buf, n + offset)
        return x

    # Uint32Array
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Uint32Array
    def Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Uint32Array
    def DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # Uint32Array
    def DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def Uint32ArrayStart(builder): builder.StartObject(1)
def Uint32ArrayAddData(builder, data): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0)
def Uint32ArrayStartDataVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def Uint32ArrayEnd(builder): return builder.EndObject()



import numpy as np
from scipy import sparse, stats
import sklearn.utils.sparsefuncs as sf


def diffexp_ttest(meanA,vA,nA,meanB,vB,nB, top_n=8, diffexp_lfc_cutoff=0.01):
    return diffexp_ttest_from_mean_var(meanA, vA, nA, meanB, vB, nB, 1000, diffexp_lfc_cutoff)


def diffexp_ttest_from_mean_var(meanA, varA, nA, meanB, varB, nB, top_n, diffexp_lfc_cutoff):
    n_var = meanA.shape[0]
    top_n = min(n_var,top_n)

    # variance / N
    vnA = varA / min(nA, nB)  # overestimate variance, would normally be nA
    vnB = varB / min(nA, nB)  # overestimate variance, would normally be nB
    sum_vn = vnA + vnB

    # degrees of freedom for Welch's t-test
    with np.errstate(divide="ignore", invalid="ignore"):
        dof = sum_vn ** 2 / (vnA ** 2 / (nA - 1) + vnB ** 2 / (nB - 1))
    dof[np.isnan(dof)] = 1

    # Welch's t-test score calculation
    with np.errstate(divide="ignore", invalid="ignore"):
        tscores = (meanA - meanB) / np.sqrt(sum_vn)
    tscores[np.isnan(tscores)] = 0

    # p-value
    pvals = stats.t.sf(np.abs(tscores), dof) * 2
    pvals_adj = pvals * n_var
    pvals_adj[pvals_adj > 1] = 1  # cap adjusted p-value at 1

    # logfoldchanges: log2(meanA / meanB)
    logfoldchanges = np.log2(np.abs((meanA + 1e-9) / (meanB + 1e-9)))

    stats_to_sort = tscores
    # find all with lfc > cutoff
    lfc_above_cutoff_idx = np.nonzero(np.abs(logfoldchanges) > diffexp_lfc_cutoff)[0]

    # derive sort order
    sort_order = np.argsort(stats_to_sort)
    sort_order = np.concatenate((sort_order[-top_n:][::-1],sort_order[:top_n][::-1]))
    if lfc_above_cutoff_idx.shape[0] > 0:
        sort_order = sort_order[np.in1d(sort_order,lfc_above_cutoff_idx)]
    # top n slice based upon sort order
    logfoldchanges_top_n = logfoldchanges[sort_order]
    pvals_top_n = pvals[sort_order]
    pvals_adj_top_n = pvals_adj[sort_order]

    # varIndex, logfoldchange, pval, pval_adj
    result = {"positive": [[sort_order[i], logfoldchanges_top_n[i], pvals_top_n[i], pvals_adj_top_n[i]] for i in
                           range(top_n)],
              "negative": [[sort_order[i], logfoldchanges_top_n[i], pvals_top_n[i], pvals_adj_top_n[i]] for i in
                           range(-1, -1 - top_n, -1)], }

    return result

def mean_var_n(X):
    """
    Two-pass variance calculation.  Numerically (more) stable
    than naive methods (and same method used by numpy.var())
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Two-pass
    """
    n = X.shape[0]
    if sparse.issparse(X):
        mean,v = sf.mean_variance_axis(X,axis=0)
    else:
        mean,v = X.mean(0),X.var(0)
    meansq = v - mean**2
    return mean, meansq, v, n



import re

import click
import requests
from requests.exceptions import ConnectionError

from .. import __version__

# Official SemVer regex: https://semver.org/
SEMVER_FORMAT = re.compile(
    r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*["
    r"a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+("
    r"?:\.[0-9a-zA-Z-]+)*))?$"
)


def log_upgrade_check():
    # Sanity-check that the CLI version is a properly-formatted SemVer string
    assert validate_version_str(__version__, release_only=False)

    # Get the current latest release
    try:
        release_tag_generator = (r["tag_name"] for r in _request_cellxgene_releases())
        latest_release = next(release_tag_generator, lambda tag_name: validate_version_str(tag_name))
        if version_gt(latest_release, __version__):
            click.echo(f"There's a new version of cellxgene available ({latest_release})!", err=True)
            click.echo("To upgrade, run the following: pip install --upgrade cellxgene\n", err=True)
    except (ConnectionError, RateLimitException):
        click.echo("Upgrade check failed.\n")


class RateLimitException(Exception):
    """
    Github API Rate Limit Exception
    """


def _request_cellxgene_releases():
    def raise_on_rate_limit(response):
        if response.status_code == 403 and res.headers.get("X-RateLimit-Remaining") == "0":
            raise RateLimitException

    url = "https://api.github.com/repos/chanzuckerberg/cellxgene/releases"
    res = requests.get(url)
    raise_on_rate_limit(res)
    for release in res.json():
        yield release
    while "next" in res.links.keys():
        res = requests.get(res.links["next"]["url"])
        raise_on_rate_limit(res)
        for release in res.json():
            yield release


def validate_version_str(version_str, release_only=True):
    """
    Test if a string conforms to SemVer format (https://semver.org/)
    :param version_str: a string to be validated
    :param release_only: only declare releases (not prereleases) valid
    :return: True if the version string is of a valid SemVer format else False
    """
    match = SEMVER_FORMAT.match(version_str)
    has_match = match is not None
    if has_match and release_only:
        return not match.group("prerelease")
    return has_match


def split_version(version_string):
    """
    Split a SemVer-formatted string into its component integers
    :param version_string: a SemVer string to be split
    :return: an array of three integers
    """
    match = SEMVER_FORMAT.match(version_string)
    return [int(match.group(group)) for group in ["major", "minor", "patch"]]


def version_gt(left_version, right_version):
    for left, right in zip(split_version(left_version), split_version(right_version)):
        if left > right:
            return True
        elif right > left:
            return False
    return False



import contextlib
import errno
import importlib.util
import logging
import os
import pkgutil
import socket
from urllib.parse import urlsplit, urljoin

import numpy as np
from flask import json

from backend.common.errors import ConfigurationError


def find_available_port(host, port=5005):
    """
    Helper method to find open port on host. Tries 5000 ports incremented from the specified port
    """
    # Takes approx 2 seconds to do a scan of 5000 ports on my laptop
    num_ports_to_try = 5000
    for port_to_try in range(port, port + num_ports_to_try):
        if is_port_available(host, port_to_try):
            return port_to_try
    raise socket.error(errno.EADDRINUSE, f"No port in range {port} - {port + num_ports_to_try - 1} available.")


def is_port_available(host, port):
    is_available = False
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        try:
            s.bind((host, port))
            is_available = True
        except socket.error:
            pass
    return is_available


def sort_options(command):
    """
    Helper for the click options - will sort options in a command, and can
    be used as a decorator.
    """
    command.params.sort(key=lambda p: p.name)
    return command


def path_join(base, *urls):
    """
    this is like urllib.parse.urljoin, except it works around the scheme-specific
    cleverness in the aforementioned code, ignores anything in the url except the path,
    and accepts more than one url.
    """
    if not base.endswith("/"):
        base += "/"
    btpl = urlsplit(base)
    path = btpl.path
    for url in urls:
        utpl = urlsplit(url)
        if btpl.scheme == "":
            path = os.path.join(path, utpl.path)
            path = os.path.normpath(path)
        else:
            path = urljoin(path, utpl.path)
    return btpl._replace(path=path).geturl()


class Float32JSONEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        """
        NaN/Infinities are illegal in standard JSON.  Python extends JSON with
        non-standard symbols that most JavaScript JSON parsers do not understand.
        The `allow_nan` parameter will force Python simplejson to throw an ValueError
        if it runs into non-finite floating point values which are unsupported by
        standard JSON.
        """
        kwargs["allow_nan"] = False
        super().__init__(*args, **kwargs)

    def default(self, obj):
        if isinstance(obj, np.float32):
            return float(obj)
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def custom_format_warning(msg, *args, **kwargs):
    return f"[cellxgene] Warning: {msg} \n"


def jsonify_numpy(data):
    return json.dumps(data, cls=Float32JSONEncoder, allow_nan=False)


def import_plugins(plugin_module):
    """
    Load optional plugin modules from server.common.plugins

    If you would like to customize cellxgene, you can add submodules to server.common.plugins before running the app.
    This code will import each, loading the code in each. If no plugins are defined, initializing the app continues as
    normal.
    """
    loaded_modules = []
    try:
        pkg = importlib.import_module(plugin_module)
        for loader, name, is_pkg in pkgutil.walk_packages(pkg.__path__):
            full_name = f"{plugin_module}.{name}"
            try:
                module = importlib.import_module(full_name)
            except Exception as e:
                raise ConfigurationError(f"Unexpected error while importing plugin: {plugin_module}.{name}: {str(e)}")
            loaded_modules.append(module)
    except ModuleNotFoundError as e:
        #  This exception occurs when the plugin_module does not exist (not an error).
        logging.debug(f"No plugins found in module: {plugin_module}: {str(e)}")

    return loaded_modules



from setuptools import setup, find_packages

with open("README.md", "rb") as fh:
    long_description = fh.read().decode()

with open("backend/server/requirements.txt") as fh:
    requirements = fh.read().splitlines()

setup(
    name="excellxgene",
    version="2.1.4",
    packages=find_packages(),
    url="https://github.com/czbiohub/excellxgene",
    license="MIT",
    author="Chan Zuckerberg Biohub",
    author_email="alexander.tarashansky@czbiohub.org",
    description="Web application for exploration of large scale scRNA-seq datasets, upgraded to enable end-to-end interactive analysis.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Framework :: Flask",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    entry_points={"console_scripts": ["excellxgene = backend.server.cli.cli:cli"]},
)



from enum import Enum


class AugmentedEnum(Enum):
    def __hash__(self):
        return self.value.__hash__()

    def __eq__(self, other):
        if isinstance(other, type(self)) or isinstance(other, str):
            return self.value == other
        return False

    def __str__(self) -> str:
        return self.value


class Axis(AugmentedEnum):
    OBS = "obs"
    VAR = "var"


class DiffExpMode(AugmentedEnum):
    TOP_N = "topN"
    VAR_FILTER = "varFilter"


JSON_NaN_to_num_warning_msg = "JSON encoding failure - please verify all data are finite values (no NaN or Infinities)"
REACTIVE_LIMIT = 1_000_000

MAX_LAYOUTS = 30



import warnings
from datetime import datetime
import anndata
import numpy as np
from packaging import version
import pandas as pd
import scipy as sp
import traceback
from pandas.core.dtypes.dtypes import CategoricalDtype
from scipy import sparse
from server_timing import Timing as ServerTiming
import time
import os
from glob import glob
import scanpy as sc
import scanpy.external as sce
from samalg import SAM
import backend.common.compute.diffexp_generic as diffexp_generic
from flask import jsonify, request, current_app, session, after_this_request, send_file
from backend.common.colors import convert_anndata_category_colors_to_cxg_category_colors
from backend.common.constants import Axis, MAX_LAYOUTS
from backend.server.common.corpora import corpora_get_props_from_anndata
from backend.common.errors import PrepareError, DatasetAccessError, FilterError
from backend.common.utils.type_conversion_utils import get_schema_type_hint_of_array
from anndata import AnnData
from backend.server.data_common.data_adaptor import DataAdaptor
from backend.common.fbs.matrix import encode_matrix_fbs
from multiprocessing import Pool
from functools import partial
import backend.server.common.rest as common_rest
import json
from backend.common.utils.utils import jsonify_numpy
import signal
import pickle
import pathlib
import base64
from hashlib import blake2b
from functools import wraps
from multiprocessing import shared_memory, resource_tracker
from os.path import exists
import sklearn.utils.sparsefuncs as sf
from numba import njit, prange, config, threading_layer
from numba.core import types
from numba.typed import Dict
#config.THREADING_LAYER = 'tbb'

global process_count
process_count = 0

anndata_version = version.parse(str(anndata.__version__)).release

def desktop_mode_only(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if current_app.hosted_mode:
            return jsonify({'message' : 'Feature only available in desktop mode.'}), 401
  
        return  f(*args, **kwargs)
  
    return decorated    

def auth0_token_required(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        token = 'profile' in session
        # return 401 if token is not passed
        if not token and current_app.hosted_mode:
            return jsonify({'message' : 'Authorization missing.'}), 401
  
        return  f(*args, **kwargs)
  
    return decorated

def anndata_version_is_pre_070():
    major = anndata_version[0]
    minor = anndata_version[1] if len(anndata_version) > 1 else 0
    return major == 0 and minor < 7

def _callback_fn(res,ws,cfn,data,post_processing,tstart):
    if post_processing is not None:
        res = post_processing(res)
    d = {"response": res,"cfn": cfn, "fail": False}
    d.update(data)
    ws.send(jsonify_numpy(d))
    global process_count
    process_count = process_count + 1
    print("Process count:",process_count,"Time elsapsed:",time.time()-tstart,"seconds")


def _multiprocessing_wrapper(da,ws,fn,cfn,data,post_processing,*args):
    _new_callback_fn = partial(_callback_fn,ws=ws,cfn=cfn,data=data,post_processing=post_processing,tstart=time.time())
    _new_error_fn = partial(_error_callback,ws=ws, cfn=cfn)
    if current_app.hosted_mode:
        da.pool.apply_async(fn,args=args, callback=_new_callback_fn, error_callback=_new_error_fn)
    else:
        try:
            res = fn(*args)
            _new_callback_fn(res)
        except Exception as e:
            _error_callback(e,ws,cfn)

def _error_callback(e, ws, cfn):
    ws.send(jsonify_numpy({"fail": True, "cfn": cfn}))
    traceback.print_exception(type(e), e, e.__traceback__)

    
def compute_diffexp_ttest(shm,shm_csc,layer,tMean,tMeanSq,obs_mask_A,obs_mask_B,top_n,lfc_cutoff):
    to_remove = []
    a,ash,ad,b,bsh,bd,c,csh,cd,Xsh = shm_csc[layer]
    to_remove.extend([a,b,c])
    shm1 = shared_memory.SharedMemory(name=a)
    shm2 = shared_memory.SharedMemory(name=b)
    shm3 = shared_memory.SharedMemory(name=c)    
    indices =np.ndarray(ash,dtype=ad,buffer=shm1.buf)
    indptr = np.ndarray(bsh,dtype=bd,buffer=shm2.buf)
    data =   np.ndarray(csh,dtype=cd,buffer=shm3.buf)
    XI = sparse.csc_matrix((data,indices,indptr),shape=Xsh)    

    iA = np.where(obs_mask_A)[0]
    iB = np.where(obs_mask_B)[0]
    niA = np.where(np.invert(np.in1d(np.arange(XI.shape[0]),iA)))[0]
    niB = np.where(np.invert(np.in1d(np.arange(XI.shape[0]),iB)))[0]
    nA = iA.size
    nB = iB.size
    if (iA.size + iB.size) == XI.shape[0]:
        n = XI.shape[0]

        if iA.size < iB.size:
            meanA,meanAsq = _partial_summer(XI.data,XI.indices,XI.indptr,XI.shape[1],iA,niA)
            meanA/=nA
            meanAsq/=nA
            vA = meanAsq - meanA**2
            vA[vA<0]=0
            meanB = (tMean*n - meanA*nA) / nB
            meanBsq = (tMeanSq*n - meanAsq*nA) / nB
            vB = meanBsq - meanB**2                          
        else:
            meanB,meanBsq = _partial_summer(XI.data,XI.indices,XI.indptr,XI.shape[1],iB,niB)
            meanB/=nB
            meanBsq/=nB
            vB = meanBsq - meanB**2
            vB[vB<0]=0
            meanA = (tMean*n - meanB*nB) / nA
            meanAsq = (tMeanSq*n - meanBsq*nB) / nA
            vA = meanAsq - meanA**2  
    else:
        meanA,meanAsq = _partial_summer(XI.data,XI.indices,XI.indptr,XI.shape[1],iA,niA)
        meanA/=nA
        meanAsq/=nA
        vA = meanAsq - meanA**2
        vA[vA<0]=0

        meanB,meanBsq = _partial_summer(XI.data,XI.indices,XI.indptr,XI.shape[1],iB,niB)
        meanB/=nB
        meanBsq/=nB
        vB = meanBsq - meanB**2
        vB[vB<0]=0   

    _unregister_shm(to_remove)        
    return diffexp_generic.diffexp_ttest(meanA,vA,nA,meanB,vB,nB,top_n,lfc_cutoff)

def save_data(shm,shm_csc,AnnDataDict,labels,labelNames,currentLayout,obs_mask,userID):
    to_remove = []

    direc = pathlib.Path().absolute()        

    fnames = glob(f"{direc}/{userID}/emb/*.p")
    embs = {}
    nnms = {}
    params={}
    pcas = {}
    for f in fnames:
        n = f.split('/')[-1][:-2]
        if exists(f) and exists(f"{direc}/{userID}/nnm/{n}.p") and exists(f"{direc}/{userID}/params/{n}.p") and exists(f"{direc}/{userID}/pca/{n}.p"):
            embs[n] = pickle.load(open(f,'rb'))
            nnms[n] = pickle.load(open(f"{direc}/{userID}/nnm/{n}.p",'rb'))
            params[n] = pickle.load(open(f"{direc}/{userID}/params/{n}.p",'rb'))
            pcas[n] = pickle.load(open(f"{direc}/{userID}/pca/{n}.p",'rb'))
        elif exists(f) and exists(f"{direc}/{userID}/nnm/{n}.p") and exists(f"{direc}/{userID}/params/{n}.p"):
            embs[n] = pickle.load(open(f,'rb'))
            nnms[n] = pickle.load(open(f"{direc}/{userID}/nnm/{n}.p",'rb'))
            params[n] = pickle.load(open(f"{direc}/{userID}/params/{n}.p",'rb'))
        elif exists(f):
            embs[n] = pickle.load(open(f,'rb'))
    
    X = embs[currentLayout]
    f = np.isnan(X).sum(1)==0    
    filt = np.logical_and(f,obs_mask)

    a,ash,ad,b,bsh,bd,c,csh,cd,Xsh = shm["X"]
    to_remove.extend([a,b,c])
    shm1 = shared_memory.SharedMemory(name=a)
    shm2 = shared_memory.SharedMemory(name=b)
    shm3 = shared_memory.SharedMemory(name=c)    
    indices =np.ndarray(ash,dtype=ad,buffer=shm1.buf)
    indptr = np.ndarray(bsh,dtype=bd,buffer=shm2.buf)
    data =   np.ndarray(csh,dtype=cd,buffer=shm3.buf)
    X = sparse.csr_matrix((data,indices,indptr),shape=Xsh)    

    adata = AnnData(X = X[filt],
            obs = AnnDataDict["obs"][filt],
            var = AnnDataDict["var"])

    for k in AnnDataDict['varm'].keys():
        adata.varm[k] = AnnDataDict['varm'][k]

    name = currentLayout.split(';')[-1]

    if labels and labelNames:
        labels = [x['__columns'][0] for x in labels]
        for n,l in zip(labelNames,labels):
            if n != "name_0":
                adata.obs[n] = pd.Categorical(l)        

    keys = list(embs.keys())
    for k in keys:
        if name not in k.split(';;'):
            del embs[k]
            if k in nnms.keys():
                del nnms[k]
            if k in params.keys():
                del params[k]
            if k in pcas.keys():
                del pcas[k]                

    temp = {}
    for key in nnms.keys():
        temp[key] = nnms[key][filt][:,filt]
    for key in temp.keys():
        adata.obsp["N_"+key] = temp[key]
    for key in params.keys():
        adata.uns["N_"+key+"_params"]=params[key]
    for key in embs.keys():
        adata.obsm["X_"+key] = embs[key][filt] 
    for key in pcas.keys():
        adata.obsm["X_"+key+"_pca"] = pcas[key][filt]         

    keys = list(adata.var.keys())
    for k in keys:
        if ";;tMean" in k:
            del adata.var[k]
                
    try:
        adata.obs_names = pd.Index(adata.obs["name_0"].astype('str'))
        del adata.obs["name_0"]
    except:
        pass
    try:
        adata.var_names = pd.Index(adata.var["name_0"].astype('str'))
        del adata.var["name_0"]
    except:
        pass        

    for k in AnnDataDict["Xs"]:
        if k != "X":
            if not (shm["X"][0] == shm["orig.X"][0] and k=="orig.X"):
                a,ash,ad,b,bsh,bd,c,csh,cd,Xsh = shm[k]
                to_remove.extend([a,b,c])
                shm1 = shared_memory.SharedMemory(name=a)
                shm2 = shared_memory.SharedMemory(name=b)
                shm3 = shared_memory.SharedMemory(name=c)    
                indices =np.ndarray(ash,dtype=ad,buffer=shm1.buf)
                indptr = np.ndarray(bsh,dtype=bd,buffer=shm2.buf)
                data =   np.ndarray(csh,dtype=cd,buffer=shm3.buf)
                X = sparse.csr_matrix((data,indices,indptr),shape=Xsh)    
                adata.layers[k] = X[filt]

    adata.write_h5ad(f"{direc}/{userID}/{userID}_{currentLayout.replace(';','_')}.h5ad")
    _unregister_shm(to_remove)
    return f"{direc}/{userID}/{userID}_{currentLayout.replace(';','_')}.h5ad"

def compute_embedding(shm,shm_csc, AnnDataDict, reembedParams, parentName, embName, userID):    
    obs_mask = AnnDataDict['obs_mask']    
    embeddingMode = reembedParams.get("embeddingMode","Preprocess and run")
    
    if embeddingMode == "Preprocess and run":
        with ServerTiming.time("layout.compute"):        
            adata = compute_preprocess(shm, shm_csc, AnnDataDict, reembedParams, userID)
            if adata.isbacked:
                raise NotImplementedError("Backed mode is incompatible with re-embedding")

            for k in list(adata.obsm.keys()):
                del adata.obsm[k]
            
            doSAM = reembedParams.get("doSAM",False)
            nTopGenesHVG = reembedParams.get("nTopGenesHVG",2000)
            nBinsHVG = reembedParams.get("nBins",20)
            doBatch = reembedParams.get("doBatch",False)
            batchMethod = reembedParams.get("batchMethod","Scanorama")
            batchKey = reembedParams.get("batchKey","")
            scanoramaKnn = reembedParams.get("scanoramaKnn",20)
            scanoramaSigma = reembedParams.get("scanoramaSigma",15)
            scanoramaAlpha = reembedParams.get("scanoramaAlpha",0.1)
            scanoramaBatchSize = reembedParams.get("scanoramaBatchSize",5000)
            bbknnNeighborsWithinBatch = reembedParams.get("bbknnNeighborsWithinBatch",3)
            numPCs = reembedParams.get("numPCs",150)
            pcaSolver = reembedParams.get("pcaSolver","randomized")
            neighborsKnn = reembedParams.get("neighborsKnn",20)
            neighborsMethod = reembedParams.get("neighborsMethod","umap")
            distanceMetric = reembedParams.get("distanceMetric","cosine")
            nnaSAM = reembedParams.get("nnaSAM",50)
            weightModeSAM = reembedParams.get("weightModeSAM","dispersion")
            umapMinDist = reembedParams.get("umapMinDist",0.1)
            scaleData = reembedParams.get("scaleData",False)
            
            if not doSAM:
                try:
                    sc.pp.highly_variable_genes(adata,flavor='seurat_v3',n_top_genes=min(nTopGenesHVG,adata.shape[1]), n_bins=nBinsHVG)                
                    adata = adata[:,adata.var['highly_variable']]                
                except:
                    print('Error during HVG selection - some of your expressions are probably negative.')
                X = adata.X
                if scaleData:
                    sc.pp.scale(adata,max_value=10)

                sc.pp.pca(adata,n_comps=min(min(adata.shape) - 1, numPCs), svd_solver=pcaSolver)
                adata.X = X
            else:            
                sam=SAM(counts = adata, inplace=True)
                X = sam.adata.X
                preprocessing = "StandardScaler" if scaleData else "Normalizer"
                sam.run(projection=None,npcs=min(min(adata.shape) - 1, numPCs), weight_mode=weightModeSAM,preprocessing=preprocessing,distance=distanceMetric,num_norm_avg=nnaSAM)
                sam.adata.X = X        
                adata=sam.adata

            if doBatch:
                if doSAM:
                    adata_batch = sam.adata
                else:
                    adata_batch = adata
                
                if batchMethod == "Harmony":
                    sce.pp.harmony_integrate(adata_batch,batchKey,adjusted_basis="X_pca")
                elif batchMethod == "BBKNN":
                    sce.pp.bbknn(adata_batch, batch_key=batchKey, metric=distanceMetric, n_pcs=numPCs, neighbors_within_batch=bbknnNeighborsWithinBatch)
                elif batchMethod == "Scanorama":
                    sce.pp.scanorama_integrate(adata_batch, batchKey, basis='X_pca', adjusted_basis='X_pca',
                                        knn=scanoramaKnn, sigma=scanoramaSigma, alpha=scanoramaAlpha,
                                        batch_size=scanoramaBatchSize)               
                if doSAM:
                    sam.adata = adata_batch
                else:
                    adata = adata_batch

            if not doSAM or doSAM and batchMethod == "BBKNN":
                if not doBatch or doBatch and batchMethod != "BBKNN":
                    sc.pp.neighbors(adata, n_neighbors=neighborsKnn, use_rep="X_pca",method=neighborsMethod, metric=distanceMetric)    
                sc.tl.umap(adata, min_dist=umapMinDist,maxiter = 500 if adata.shape[0] <= 10000 else 200)
            else:
                sam.run_umap(metric=distanceMetric,min_dist=umapMinDist)
                adata.obsm['X_umap'] = sam.adata.obsm['X_umap']
                adata.obsp['connectivities'] = sam.adata.obsp['connectivities']
                
            umap = adata.obsm["X_umap"]
            result = np.full((obs_mask.shape[0], umap.shape[1]), np.NaN)
            result[obs_mask] = umap
            X_umap,nnm = result, adata.obsp['connectivities'] 
            obsm = adata.obsm['X_pca']
            pca = np.full((obs_mask.shape[0], obsm.shape[1]), np.NaN)
            pca[obs_mask] = obsm
                    
    elif embeddingMode == "Create embedding from subset":
        direc = pathlib.Path().absolute()    
        umap = pickle.load(open(f"{direc}/{userID}/emb/{parentName}.p",'rb'))                     
        result = np.full((obs_mask.shape[0], umap.shape[1]), np.NaN)
        result[obs_mask] = umap[obs_mask] 
        X_umap = result  
        nnm = pickle.load(open(f"{direc}/{userID}/nnm/{parentName}.p",'rb'))[obs_mask][:,obs_mask]

        try:
            obsm = pickle.load(open(f"{direc}/{userID}/pca/{parentName}.p",'rb'))
        except:
            obsm = np.zeros(X_umap.shape)
        pca = np.full((obs_mask.shape[0], obsm.shape[1]), np.NaN)
        pca[obs_mask] = obsm[obs_mask]       

    elif embeddingMode == "Run UMAP":
        doBatch = reembedParams.get("doBatch",False)
        batchMethod = reembedParams.get("batchMethod","Scanorama")
        batchKey = reembedParams.get("batchKey","")
        scanoramaKnn = reembedParams.get("scanoramaKnn",20)
        scanoramaSigma = reembedParams.get("scanoramaSigma",15)
        scanoramaAlpha = reembedParams.get("scanoramaAlpha",0.1)
        scanoramaBatchSize = reembedParams.get("scanoramaBatchSize",5000)
        bbknnNeighborsWithinBatch = reembedParams.get("bbknnNeighborsWithinBatch",3)
        neighborsKnn = reembedParams.get("neighborsKnn",20)
        neighborsMethod = reembedParams.get("neighborsMethod","umap")
        distanceMetric = reembedParams.get("distanceMetric","cosine")
        umapMinDist = reembedParams.get("umapMinDist",0.1)
        latentSpace = reembedParams.get("latentSpace","")


        direc = pathlib.Path().absolute()    
        n = latentSpace.split("X_")[-1]
        obsm = pickle.load(open(f"{direc}/{userID}/emb/{n}.p",'rb'))   
        adata = AnnData(X=np.zeros(obsm.shape)[obs_mask],obsm={"X_pca":obsm[obs_mask]})    

        if doBatch:
            if batchMethod == "Harmony":
                sce.pp.harmony_integrate(adata,batchKey,adjusted_basis="X_pca")
            elif batchMethod == "BBKNN":
                sce.pp.bbknn(adata, batch_key=batchKey, metric=distanceMetric, n_pcs=obsm.shape[1], neighbors_within_batch=bbknnNeighborsWithinBatch)
            elif batchMethod == "Scanorama":
                sce.pp.scanorama_integrate(adata, batchKey, basis='X_pca', adjusted_basis='X_pca',
                                    knn=scanoramaKnn, sigma=scanoramaSigma, alpha=scanoramaAlpha,
                                    batch_size=scanoramaBatchSize)               
        
        if not doBatch or doBatch and batchMethod != "BBKNN":
            sc.pp.neighbors(adata, n_neighbors=neighborsKnn, use_rep="X_pca",method=neighborsMethod, metric=distanceMetric)    
        sc.tl.umap(adata, min_dist=umapMinDist,maxiter = 500 if adata.shape[0] <= 10000 else 200)
        umap = adata.obsm["X_umap"]
        nnm = adata.obsp["connectivities"]
        result = np.full((obs_mask.shape[0], umap.shape[1]), np.NaN)
        result[obs_mask] = umap 
        X_umap = result
        
        obsm = adata.obsm['X_pca']
        pca = np.full((obs_mask.shape[0], obsm.shape[1]), np.NaN)
        pca[obs_mask] = obsm        

    if embName == "":
        embName = f"umap_{str(hex(int(time.time())))[2:]}"

    if parentName != "":
        parentName+=";;"
    
    name = f"{parentName}{embName}"
    if exists(f"{userID}/emb/{name}.p"):
        name = f"{name}_{str(hex(int(time.time())))[2:]}"
        
    dims = [f"{name}_0", f"{name}_1"]
    layout_schema = {"name": name, "type": "float32", "dims": dims}

    IXer = pd.Series(index =np.arange(nnm.shape[0]), data = np.where(obs_mask.flatten())[0])
    x,y = nnm.nonzero()
    d = nnm.data
    nnm = sp.sparse.coo_matrix((d,(IXer[x].values,IXer[y].values)),shape=(obs_mask.size,)*2).tocsr()

    direc = pathlib.Path().absolute()        
    if exists(f"{direc}/{userID}/params/latest.p"):
        latestPreParams = pickle.load(open(f"{direc}/{userID}/params/latest.p","rb"))
    else:
        latestPreParams = None

    if exists(f"{userID}/params/{parentName}.p"):
        parentParams = pickle.load(open(f"{direc}/{userID}/params/{parentName}.p","rb"))
    else:
        parentParams = None

    if latestPreParams is not None:
        for k in latestPreParams.keys():
            reembedParams[k] = latestPreParams[k]
            
    if (parentParams is not None):
        reembedParams[f"parentParams"]=parentParams

    pickle.dump(nnm, open(f"{direc}/{userID}/nnm/{name}.p","wb"))
    pickle.dump(X_umap, open(f"{direc}/{userID}/emb/{name}.p","wb"))
    pickle.dump(reembedParams, open(f"{direc}/{userID}/params/{name}.p","wb"))
    pickle.dump(pca, open(f"{direc}/{userID}/pca/{name}.p","wb"))

    return layout_schema

def compute_leiden(obs_mask,name,resolution,userID):
    direc = pathlib.Path().absolute() 
    nnm = pickle.load(open(f"{direc}/{userID}/nnm/{name}.p","rb"))            
    nnm = nnm[obs_mask][:,obs_mask]

    X = nnm

    import igraph as ig
    import leidenalg

    adjacency = X
    sources, targets = adjacency.nonzero()
    weights = adjacency[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph(directed=True)
    g.add_vertices(adjacency.shape[0])
    g.add_edges(list(zip(sources, targets)))
    try:
        g.es["weight"] = weights
    except BaseException:
        pass

    cl = leidenalg.find_partition(
        g, leidenalg.RBConfigurationVertexPartition, resolution_parameter=resolution,seed=0
    )
    result = np.array(cl.membership)
    clusters = np.array(["unassigned"]*obs_mask.size,dtype='object')
    clusters[obs_mask] = result.astype('str')
    return list(result)

def compute_sankey_df(labels, name, obs_mask, userID):
    def reducer(a, b):
        result_a, inv_ndx = np.unique(a, return_inverse=True)
        result_b = np.bincount(inv_ndx, weights=b)
        return result_a, result_b        
    def cantor(a,b):
        return ((a+b)*(a+b+1)/2+b).astype('int')
    def inv_cantor(z):
        w = np.floor((np.sqrt(8*z + 1) - 1)/2)
        t = (w**2 + w)/2
        y = (z-t).astype('int')
        x = (w-y).astype('int')
        return x,y
    
    direc = pathlib.Path().absolute() 
    nnm = pickle.load(open(f"{direc}/{userID}/nnm/{name}.p","rb"))              
    nnm = nnm[obs_mask][:,obs_mask]

    cl=[]
    clu = []
    rixers=[]
    unassigned_ints=[]
    for i,c in enumerate(labels):
        cl0 = np.array(['A'+str(i)+'_'+str(x).replace(' ','_').replace('(','_').replace(')','_') for x in c])
        clu0,cluc0 = np.unique(cl0,return_counts=True)
        ix = pd.Series(index=clu0,data=np.arange(clu0.size))
        cl0 = ix[cl0].values
        ll = np.arange(clu0.size)[clu0=="A"+str(i)+"_unassigned"]
        if ll.size > 0:
            unassigned_ints.append(ll[0])
        else:
            unassigned_ints.append(-1)
            
        rixers.append(pd.Series(data=clu0,index=np.arange(clu0.size)))                     
        clu0 = np.arange(clu0.size)
        clu.append((clu0,cluc0))
        cl.append(cl0)

    ps = []
    cs = []
    for i,cl1 in enumerate(cl[:-1]):
        j = i+1
        cl2 = cl[i+1]
        clu1,cluc1 = clu[i]
        clu2,cluc2 = clu[j]
        uint1 = unassigned_ints[i]
        uint2 = unassigned_ints[j]
        rixer1 = rixers[i]
        rixer2 = rixers[j]        
        
        ac = pd.Series(index=clu1,data=cluc1)
        bc = pd.Series(index=clu2,data=cluc2)

        ixer1 = pd.Series(data=np.arange(clu1.size),index=clu1)
        ixer2 = pd.Series(data=np.arange(clu2.size),index=clu2)

        xi,yi = nnm.nonzero()
        di = nnm.data

        px,py = cl1[xi],cl2[yi]
        filt = np.logical_and(px != uint1,py != uint2)
        px = px[filt]
        py = py[filt]
        dif = di[filt]

        p = cantor(px,py)

        keys,cluster_scores = reducer(p,dif)
        xc,yc = inv_cantor(keys)
        cluster_scores = cluster_scores / ac[xc].values

        xc=ixer1[xc].values
        yc=ixer2[yc].values

        CSIM = sp.sparse.coo_matrix((cluster_scores,(xc,yc)),shape=(clu1.size,clu2.size)).A


        xi,yi = nnm.nonzero()
        di = nnm.data

        px,py = cl2[xi],cl1[yi]
        filt = np.logical_and(px != uint2,py != uint1)
        px = px[filt]
        py = py[filt]
        dif = di[filt]

        p = cantor(px,py)

        keys,cluster_scores = reducer(p,dif)
        xc,yc = inv_cantor(keys)
        cluster_scores = cluster_scores / bc[xc].values


        xc=ixer2[xc].values
        yc=ixer1[yc].values

        CSIM2 = sp.sparse.coo_matrix((cluster_scores,(xc,yc)),shape=(clu2.size,clu1.size)).A


        CSIM = np.stack((CSIM,CSIM2.T),axis=2).min(2)
        x,y = CSIM.nonzero()
        d = CSIM[x,y]
        x,y = rixer1[clu1[x]].values,rixer2[clu2[y]].values
        ps.append(np.vstack((x,y)).T)
        cs.append(d)

    ps = np.vstack(ps)
    cs = np.concatenate(cs)
    ps = [list(x) for x in ps]
    cs = list(cs)        
    return {"edges":ps,"weights":cs}

def compute_preprocess(shm,shm_csc, AnnDataDict, reembedParams, userID):
    to_remove = []
    layers = AnnDataDict['Xs'] 
    obs = AnnDataDict['obs']
    root = AnnDataDict['X_root']
    obs_mask = AnnDataDict['obs_mask']
    kkk=layers[0]
    a,ash,ad,b,bsh,bd,c,csh,cd,Xsh = shm[kkk]
    to_remove.extend([a,b,c])
    shm1 = shared_memory.SharedMemory(name=a)
    shm2 = shared_memory.SharedMemory(name=b)
    shm3 = shared_memory.SharedMemory(name=c)    
    indices =np.ndarray(ash,dtype=ad,buffer=shm1.buf)
    indptr = np.ndarray(bsh,dtype=bd,buffer=shm2.buf)
    data =   np.ndarray(csh,dtype=cd,buffer=shm3.buf)
    X = sparse.csr_matrix((data,indices,indptr),shape=Xsh)[obs_mask]
    adata = AnnData(X=X,obs=obs[obs_mask])

    adata.layers[layers[0]] = X
    for k in layers[1:]:
        kkk=k
        a,ash,ad,b,bsh,bd,c,csh,cd,Xsh = shm[kkk]
        to_remove.extend([a,b,c])
        shm1 = shared_memory.SharedMemory(name=a)
        shm2 = shared_memory.SharedMemory(name=b)
        shm3 = shared_memory.SharedMemory(name=c)    
        indices =np.ndarray(ash,dtype=ad,buffer=shm1.buf)
        indptr = np.ndarray(bsh,dtype=bd,buffer=shm2.buf)
        data =   np.ndarray(csh,dtype=cd,buffer=shm3.buf)
        X = sparse.csr_matrix((data,indices,indptr),shape=Xsh)[obs_mask]       
        adata.layers[k] = X
    adata.obsm["X_root"] = root[obs_mask]

    doBatchPrep = reembedParams.get("doBatchPrep",False)
    batchPrepParams = reembedParams.get("batchPrepParams",{})
    batchPrepKey = reembedParams.get("batchPrepKey","")
    batchPrepLabel = reembedParams.get("batchPrepLabel","")

    doPreprocess = reembedParams.get("doPreprocess",False)
    minCountsCF = reembedParams.get("minCountsCF",0)
    minGenesCF = reembedParams.get("minGenesCF",0)
    minCellsGF = reembedParams.get("minCellsGF",0)
    maxCellsGF = reembedParams.get("maxCellsGF",100)
    minCountsGF = reembedParams.get("minCountsGF",0)
    logTransform = reembedParams.get("logTransform",False)
    dataLayer = reembedParams.get("dataLayer","X")
    sumNormalizeCells = reembedParams.get("sumNormalizeCells",False)
    
    cn = np.array(list(adata.obs["name_0"]))        
    filt = np.array([True]*adata.shape[0])
    if doBatchPrep and batchPrepKey != "" and batchPrepLabel != "":
        cl = np.array(list(adata.obs[batchPrepKey]))
        batches = np.unique(cl)
        adatas = []
        cns = []
        for k in batches:
            params = batchPrepParams[batchPrepKey].get(k,{})

            doPreprocess = params.get("doPreprocess",False)
            minCountsCF = params.get("minCountsCF",0)
            minGenesCF = params.get("minGenesCF",0)
            minCellsGF = params.get("minCellsGF",0)
            maxCellsGF = params.get("maxCellsGF",100)
            minCountsGF = params.get("minCountsGF",0)
            logTransform = params.get("logTransform",False)
            dataLayer = params.get("dataLayer","X")
            sumNormalizeCells = params.get("sumNormalizeCells",False)
            
            adata_sub = adata[cl==k].copy()
            adata_sub.obs_names = adata_sub.obs["name_0"]
            if dataLayer == "X":
                adata_sub_raw = adata_sub
                if dataLayer == "X" and "X" not in adata_sub_raw.layers.keys():
                    adata_sub_raw.layers["X"] = adata_sub_raw.X    
                adata_sub_raw.X = adata_sub_raw.layers[dataLayer]        
            else:
                adata_sub_raw = AnnData(X=adata_sub.layers[dataLayer])
                adata_sub_raw.var_names = adata_sub.var_names
                adata_sub_raw.obs_names = adata_sub.obs_names
                adata_sub_raw.obs = adata_sub.obs
                for key in adata_sub.var.keys():
                    adata_sub_raw.var[key] = adata_sub.var[key]   
            if doPreprocess:
                filt1,_ = sc.pp.filter_cells(adata_sub_raw,min_counts=minCountsCF, inplace=False)
                filt2,_ = sc.pp.filter_cells(adata_sub_raw,min_genes=minGenesCF, inplace=False)
                filt = np.logical_and(filt1,filt2)
                cns.extend(np.array(list(adata_sub_raw.obs["name_0"]))[filt])
                target_sum = np.median(np.array(adata_sub_raw.X[filt].sum(1)).flatten())
                a1,_=sc.pp.filter_genes(adata_sub_raw, min_counts=minCountsGF,inplace=False)
                a2,_=sc.pp.filter_genes(adata_sub_raw, min_cells=minCellsGF/100*adata_sub_raw.shape[0],inplace=False)
                a3,_=sc.pp.filter_genes(adata_sub_raw, max_cells=maxCellsGF/100*adata_sub_raw.shape[0],inplace=False)
                a = a1*a2*a3
                
                adata_sub_raw.X = adata_sub_raw.X.multiply(a.flatten()[None,:]).tocsr()

                if sumNormalizeCells:
                    sc.pp.normalize_total(adata_sub_raw,target_sum=target_sum)
                if logTransform:
                    try:
                        sc.pp.log1p(adata_sub_raw)  
                    except:
                        pass
            else: 
                cns.extend(np.array(list(adata_sub_raw.obs["name_0"])))

            adatas.append(adata_sub_raw)
        adata_raw = anndata.concat(adatas,axis=0,join="inner")
        filt = np.in1d(np.array(list(cn)),np.array(cns))
        temp = adata_raw.obs_names.copy()
        adata_raw.obs_names = adata_raw.obs["name_0"]
        adata_raw = adata_raw[cn]
        adata_raw.obs_names = temp
    else:
        if dataLayer == "X":
            adata_raw = adata.copy()
            if dataLayer == "X" and "X" not in adata_raw.layers.keys():
                adata_raw.layers["X"] = adata_raw.X    
            adata_raw.X = adata_raw.layers[dataLayer]        
        else:

            adata_raw = AnnData(X=adata.layers[dataLayer])
            adata_raw.var_names = adata.var_names
            adata_raw.obs_names = adata.obs_names
            adata_raw.obs = adata.obs
            for key in adata.var.keys():
                adata_raw.var[key] = adata.var[key]                
        if doPreprocess:
            filt1,_ = sc.pp.filter_cells(adata_raw,min_counts=minCountsCF, inplace=False)
            filt2,_ = sc.pp.filter_cells(adata_raw,min_genes=minGenesCF, inplace=False)
            filt = np.logical_and(filt1,filt2)
            target_sum = np.median(np.array(adata_raw.X[filt].sum(1)).flatten())
            a1,_=sc.pp.filter_genes(adata_raw, min_counts=minCountsGF,inplace=False)
            a2,_=sc.pp.filter_genes(adata_raw, min_cells=minCellsGF/100*adata_raw.shape[0],inplace=False)
            a3,_=sc.pp.filter_genes(adata_raw, max_cells=maxCellsGF/100*adata_raw.shape[0],inplace=False)
            a = a1*a2*a3
            
            adata_raw.X = adata_raw.X.multiply(a.flatten()[None,:]).tocsr()
            
            if sumNormalizeCells:
                sc.pp.normalize_total(adata_raw,target_sum=target_sum)
            if logTransform:
                try:
                    sc.pp.log1p(adata_raw) 
                except:
                    pass
        
    direc = pathlib.Path().absolute() 
   
    adata_raw.layers['X'] = adata_raw.X            
    doBatchPrep = reembedParams.get("doBatchPrep",False)
    batchPrepParams = reembedParams.get("batchPrepParams",{})
    batchPrepKey = reembedParams.get("batchPrepKey","")
    batchPrepLabel = reembedParams.get("batchPrepLabel","")

    doPreprocess = reembedParams.get("doPreprocess",False)
    minCountsCF = reembedParams.get("minCountsCF",0)
    minGenesCF = reembedParams.get("minGenesCF",0)
    minCellsGF = reembedParams.get("minCellsGF",0)
    maxCellsGF = reembedParams.get("maxCellsGF",100)
    minCountsGF = reembedParams.get("minCountsGF",0)
    logTransform = reembedParams.get("logTransform",False)
    dataLayer = reembedParams.get("dataLayer","X")
    sumNormalizeCells = reembedParams.get("sumNormalizeCells",False)

    prepParams = {
        "doBatchPrep":doBatchPrep,
        "batchPrepParams":batchPrepParams,
        "batchPrepKey":batchPrepKey,
        "batchPrepLabel":batchPrepLabel,
        "doPreprocess":doPreprocess,
        "minCountsCF":minCountsCF,
        "minGenesCF":minGenesCF,
        "minCellsGF":minCellsGF,
        "maxCellsGF":maxCellsGF,
        "minCountsGF":minCountsGF,
        "logTransform":logTransform,
        "dataLayer":dataLayer,
        "sumNormalizeCells":sumNormalizeCells,        
    }        
    pickle.dump(prepParams, open(f"{direc}/{userID}/params/latest.p","wb"))  
    _unregister_shm(to_remove) 
    return adata_raw

def _unregister_shm(to_remove):
    to_remove = list(np.unique(to_remove))
    already_deleted=[]
    for s in to_remove:
        if s not in already_deleted:
            resource_tracker.unregister("/"+s,"shared_memory")      
            already_deleted.append(s)     

def initialize_socket(da):
    sock = da.socket
    @sock.route("/diffexp")
    @auth0_token_required
    def diffexp(ws):
        while True:
            data = ws.receive()
            if data is not None:  
                data = json.loads(data)
                obsFilterA = data.get("set1", {"filter": {}})["filter"]
                obsFilterB = data.get("set2", {"filter": {}})["filter"]
                layer = data.get("layer","X")
                top_n = data.get("count", 100)
                lfc_cutoff = 0.01
                shape = da.get_shape()

                obs_mask_A = da._axis_filter_to_mask(Axis.OBS, obsFilterA["obs"], shape[0])
                obs_mask_B = da._axis_filter_to_mask(Axis.OBS, obsFilterB["obs"], shape[0])      

                tMean = da.data.var[f'{layer};;tMean'].values
                tMeanSq = da.data.var[f'{layer};;tMeanSq'].values                      

                _multiprocessing_wrapper(da,ws,compute_diffexp_ttest, "diffexp",data,None,da.shm_layers_csr,da.shm_layers_csc,layer,tMean,tMeanSq,obs_mask_A,obs_mask_B,top_n,lfc_cutoff)
    
    @sock.route("/reembedding")
    @auth0_token_required
    def reembedding(ws):
        while True:
            data = ws.receive()
            if data is not None:  
                data = json.loads(data)

                filter = data["filter"]
                if len(filter["obs"]["index"]) <= 50000: 
                    reembedParams = data["params"] if data else {}
                    parentName = data["parentName"] if data else ""
                    embName = data["embName"] if data else None
        
                    annotations = da.dataset_config.user_annotations        
                    userID = f"{annotations._get_userdata_idhash(da)}"  
                    layers = []
                    if current_app.hosted_mode:
                        doBatchPrep = reembedParams.get("doBatchPrep",False)
                        batchPrepParams = reembedParams.get("batchPrepParams",{})
                        batchPrepKey = reembedParams.get("batchPrepKey","")
                        batchPrepLabel = reembedParams.get("batchPrepLabel","")
                        dataLayer = reembedParams.get("dataLayer","X")
                        if doBatchPrep and batchPrepKey != "" and batchPrepLabel != "":
                            cl = np.array(list(da.data.obs[batchPrepKey]))
                            batches = np.unique(cl)
                            for k in batches:
                                params = batchPrepParams[batchPrepKey].get(k,{})
                                k = params.get("dataLayer","X")
                                layers.append(k)
                        else:
                            layers.append(dataLayer)
                    else:
                        dataLayer = reembedParams.get("dataLayer","X")
                        layers.append(dataLayer)
                    layers = list(np.unique(layers))
                    direc = pathlib.Path().absolute()  
                    obs = pickle.load(open(f"{direc}/{userID}/obs.p",'rb'))

                    obs['name_0'] = obs.index
                    obs.index = pd.Index(np.arange(obs.shape[0]))
                    AnnDataDict = {
                        "Xs": layers,
                        "obs": obs,
                        "X_root":da._obsm_init["X_root"],
                        "obs_mask": da._axis_filter_to_mask(Axis.OBS, filter["obs"], da.get_shape()[0])
                    }

                    def post_processing(res):
                        da.schema["layout"]["obs"].append(res)
                        return res
        
                    _multiprocessing_wrapper(da,ws,compute_embedding, "reembedding",data,post_processing,da.shm_layers_csr,da.shm_layers_csc,AnnDataDict, reembedParams, parentName, embName, userID)
    """
    @sock.route("/preprocessing")
    @desktop_mode_only
    def preprocessing(ws):
        while True:
            data = ws.receive()
            if data is not None and not current_app.hosted_mode:  
                data = json.loads(data)

                filter = data["filter"] if data else None
                reembedParams = data["params"] if data else {}
    
                annotations = da.dataset_config.user_annotations        
                userID = f"{annotations._get_userdata_idhash(da)}"  

                layers = []
                doBatchPrep = reembedParams.get("doBatchPrep",False)
                batchPrepParams = reembedParams.get("batchPrepParams",{})
                batchPrepKey = reembedParams.get("batchPrepKey","")
                batchPrepLabel = reembedParams.get("batchPrepLabel","")
                dataLayer = reembedParams.get("dataLayer","X")
                if doBatchPrep and batchPrepKey != "" and batchPrepLabel != "":
                    cl = np.array(list(da.data.obs[batchPrepKey]))
                    batches = np.unique(cl)
                    for k in batches:
                        params = batchPrepParams[batchPrepKey].get(k,{})
                        k = params.get("dataLayer","X")
                        layers.append(k)
                else:
                    layers.append(dataLayer)
                layers = list(np.unique(layers))
                
                AnnDataDict = {
                    "Xs":layers,
                    "obs":da.data.obs,
                    "X_root":da._obsm_init["X_root"],
                    "obs_mask": da._axis_filter_to_mask(Axis.OBS, filter["obs"], da.get_shape()[0])
                }

                def post_processing(res):
                    if da.shm_layers_csr["X"][0] != da.shm_layers_csr["orig.X"][0]:
                        for j in [0,3,6]:
                            s = shared_memory.SharedMemory(name=da.shm_layers_csr["X"][j])
                            s.close()
                            s.unlink()     
                    if da.shm_layers_csc["X"][0] != da.shm_layers_csc["orig.X"][0]:
                        for j in [0,3,6]:
                            s = shared_memory.SharedMemory(name=da.shm_layers_csc["X"][j])
                            s.close()
                            s.unlink()                              
                    da.shm_layers_csr["X"] = res["X_shm"]                    
                    da.shm_layers_csc["X"] = res["X_shm_csc"]

                    da.data.X = res["X"]
                    da.data.layers["X"] = res["X"]
                    da.data.var["X;;tMean"] = res['mean']
                    da.data.var["X;;tMeanSq"] = res['meansq']
                    return da.get_schema()

                _multiprocessing_wrapper(da,ws,compute_preprocess, "preprocessing",data,post_processing,da.shm_layers_csr,da.shm_layers_csc,AnnDataDict,reembedParams,userID, False)
    """
    @sock.route("/sankey")
    def sankey(ws):
        while True:
            data = ws.receive()
            if data is not None:  
                data = json.loads(data)
                labels = data.get("labels", None)
                name = data.get("name", None)
                filter = data.get("filter",None)
                annotations = da.dataset_config.user_annotations        
                userID = f"{annotations._get_userdata_idhash(da)}"  

                obs_mask = da._axis_filter_to_mask(Axis.OBS, filter["obs"], da.get_shape()[0])
                _multiprocessing_wrapper(da,ws,compute_sankey_df, "sankey",data,None,labels, name, obs_mask, userID)              

    @sock.route("/downloadAnndata")
    @auth0_token_required
    def downloadAnndata(ws):
        while True:
            data = ws.receive()
            if data is not None:  
                data = json.loads(data)
                labels = data.get("labels",None)
                labelNames = data.get("labelNames",None)
                currentLayout = data.get("currentLayout",None)
                filter = data["filter"] if data else None

                shape = da.get_shape()
                obs_mask = da._axis_filter_to_mask(Axis.OBS, filter["obs"], shape[0])

                annotations = da.dataset_config.user_annotations        
                userID = f"{annotations._get_userdata_idhash(da)}" 
                
                layers = list(da.data.layers.keys())
                varm = {}
                for k in da.data.varm.keys():
                    varm[k] = da.data.varm[k]

                AnnDataDict={"Xs":layers,"obs":da.data.obs, "var": da.data.var, "varm": varm}

                _multiprocessing_wrapper(da,ws,save_data, "downloadAnndata",data,None,da.shm_layers_csr,da.shm_layers_csc,AnnDataDict,labels,labelNames,currentLayout,obs_mask,userID)

 
    @sock.route("/leiden")
    @auth0_token_required
    def leiden(ws):
        while True:
            data = ws.receive()
            if data is not None:  
                data = json.loads(data)
                name = data.get("name", None)
                resolution = data.get('resolution',1.0)
                filter = data.get('filter',None)
                annotations = da.dataset_config.user_annotations        
                userID = f"{annotations._get_userdata_idhash(da)}"  
                obs_mask = da._axis_filter_to_mask(Axis.OBS, filter["obs"], da.get_shape()[0])

                _multiprocessing_wrapper(da,ws,compute_leiden, "leiden",data,None,obs_mask,name,resolution,userID)

@njit(parallel=True)
def _partial_summer(d,x,ptr,m,inc,ninc, calculate_sq=True):
    htable = Dict.empty(
        key_type=types.int64,
        value_type=types.boolean,
    )    
    for i in inc:
        htable[i] = True
    
    for i in ninc:
        htable[i] = False
        
    res = np.zeros(m)
    res2 = np.zeros(m)
    for i in prange(m):
        di = d[ptr[i] : ptr[i+1]]
        xi = x[ptr[i] : ptr[i+1]]
        s=0
        if calculate_sq:
            s2 = 0
        for j in prange(xi.size):
            s += di[j] if htable[xi[j]] else 0
            if calculate_sq:
                s2 += di[j]**2 if htable[xi[j]] else 0
                
        res[i] = s
        if calculate_sq:
            res2[i] = s2
    return res,res2

@njit(parallel=True)
def _fmt_swapper(indices,indptr,data,n):#x,y,d,ptr):
    pair = np.zeros_like(indices)
    for i in prange(indptr.size):
        pair[indptr[i]:indptr[i+1]]=i

    indptr2 = np.zeros(n,dtype=indices.dtype)
    for i in range(indices.size):
        indptr2[indices[i]+1]+=1
    indptr2 = np.cumsum(indptr2)

    res = np.zeros_like(pair)
    dres = np.zeros_like(data)
    indptr3 = indptr2[:-1].copy()  
    for i in range(indices.size):
        j = indices[i]
        k = indptr3[j]
        res[k]=pair[i]
        dres[k]=data[i]
        indptr3[j]+=1
    return dres,res,indptr2

def fmt_swapper(X):
    import scipy as sp
    if X.getformat()=="csc":
        return sp.sparse.csr_matrix(_fmt_swapper(X.indices,X.indptr,X.data,X.shape[0]+1),shape=X.shape)
    elif X.getformat()=="csr":
        return sp.sparse.csc_matrix(_fmt_swapper(X.indices,X.indptr,X.data,X.shape[1]+1),shape=X.shape)

def _create_shm(X):
    shm = shared_memory.SharedMemory(create=True,size=X.nbytes)
    a = np.ndarray(X.shape, dtype = X.dtype, buffer = shm.buf)
    a[:] = X[:]
    return shm.name

def _create_shm_from_data(X):
    a = _create_shm(X.indices)
    b = _create_shm(X.indptr)
    c = _create_shm(X.data)    
    return (a,X.indices.shape,X.indices.dtype,b,X.indptr.shape,X.indptr.dtype,c,X.data.shape,X.data.dtype,X.shape)

"""
def _create_data_from_shm(a,ash,ad,b,bsh,bd,c,csh,cd,Xsh):
    import scipy as sp
    shm1 = shared_memory.SharedMemory(name=a)
    shm2 = shared_memory.SharedMemory(name=b)
    shm3 = shared_memory.SharedMemory(name=c)    
    indices =np.ndarray(ash,dtype=ad,buffer=shm1.buf)
    indptr = np.ndarray(bsh,dtype=bd,buffer=shm2.buf)
    data =   np.ndarray(csh,dtype=cd,buffer=shm3.buf)
    return sp.sparse.csr_matrix((data,indices,indptr),shape=Xsh)

def create_data_from_shm(r):
    return _create_data_from_shm(*r)
"""

def _initializer():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

class AnndataAdaptor(DataAdaptor):

    def __init__(self, data_locator, app_config=None, dataset_config=None):
        super().__init__(data_locator, app_config, dataset_config)
        self.data = None
        if app_config.hosted_mode:
            self._create_pool()

        self._load_data(data_locator, root_embedding=app_config.root_embedding)    
        self._validate_and_initialize()

    def _create_pool(self):
        self.pool = Pool(os.cpu_count(), initializer=_initializer, maxtasksperchild=None)

    def cleanup(self):
        pass

    @staticmethod
    def pre_load_validation(data_locator):
        if data_locator.islocal():
            # if data locator is local, apply file system conventions and other "cheap"
            # validation checks.  If a URI, defer until we actually fetch the data and
            # try to read it.  Many of these tests don't make sense for URIs (eg, extension-
            # based typing).
            if not data_locator.exists():
                raise DatasetAccessError("does not exist")

    @staticmethod
    def file_size(data_locator):
        return data_locator.size() if data_locator.islocal() else 0

    @staticmethod
    def open(data_locator, app_config, dataset_config=None):
        return AnndataAdaptor(data_locator, app_config, dataset_config)

    def get_corpora_props(self):
        return corpora_get_props_from_anndata(self.data)

    def get_name(self):
        return "cellxgene anndata adaptor version"

    def get_library_versions(self):
        return dict(anndata=str(anndata.__version__))

    @staticmethod
    def _create_unique_column_name(df, col_name_prefix):
        """given the columns of a dataframe, and a name prefix, return a column name which
        does not exist in the dataframe, AND which is prefixed by `prefix`

        The approach is to append a numeric suffix, starting at zero and increasing by
        one, until an unused name is found (eg, prefix_0, prefix_1, ...).
        """
        suffix = 0
        while f"{col_name_prefix}{suffix}" in df:
            suffix += 1
        return f"{col_name_prefix}{suffix}"

    def compute_diffexp_ttest(self):
        pass

    def compute_embedding(self):
        pass

    def compute_sankey_df(self):
        pass

    def compute_leiden(self):
        pass

    def _alias_annotation_names(self):
        """
        The front-end relies on the existance of a unique, human-readable
        index for obs & var (eg, var is typically gene name, obs the cell name).
        The user can specify these via the --obs-names and --var-names config.
        If they are not specified, use the existing index to create them, giving
        the resulting column a unique name (eg, "name").

        In both cases, enforce that the result is unique, and communicate the
        index column name to the front-end via the obs_names and var_names config
        (which is incorporated into the schema).
        """
        self.original_obs_index = self.data.obs.index

        for (ax_name, var_name) in ((Axis.OBS, "obs"), (Axis.VAR, "var")):
            config_name = f"single_dataset__{var_name}_names"
            parameter_name = f"{var_name}_names"
            name = getattr(self.server_config, config_name)
            df_axis = getattr(self.data, str(ax_name))
            if name is None:
                # Default: create unique names from index
                if not df_axis.index.is_unique:
                    raise KeyError(
                        f"Values in {ax_name}.index must be unique. "
                        "Please prepare data to contain unique index values, or specify an "
                        "alternative with --{ax_name}-name."
                    )
                name = self._create_unique_column_name(df_axis.columns, "name_")
                self.parameters[parameter_name] = name
                # reset index to simple range; alias name to point at the
                # previously specified index.
                df_axis.rename_axis(name, inplace=True)
                df_axis.reset_index(inplace=True)
            elif name in df_axis.columns:
                # User has specified alternative column for unique names, and it exists
                if not df_axis[name].is_unique:
                    raise KeyError(
                        f"Values in {ax_name}.{name} must be unique. " "Please prepare data to contain unique values."
                    )
                df_axis.reset_index(drop=True, inplace=True)
                self.parameters[parameter_name] = name
            else:
                # user specified a non-existent column name
                raise KeyError(f"Annotation name {name}, specified in --{ax_name}-name does not exist.")

    def _create_schema(self):
        layers = list(self.data.layers.keys())
        
        if "X" not in layers:
            layers = ["X"] + layers
        
        self.schema = {
            "dataframe": {"nObs": self.cell_count, "nVar": self.gene_count, "type": str(self.data.X.dtype)},
            "annotations": {
                "obs": {"index": self.parameters.get("obs_names"), "columns": []},
                "var": {"index": self.parameters.get("var_names"), "columns": []},
            },
            "layout": {"obs": []},
            "layers": layers
        }
        for ax in Axis:
            curr_axis = getattr(self.data, str(ax))
            for ann in curr_axis:
                ann_schema = {"name": ann, "writable": True}
                ann_schema.update(get_schema_type_hint_of_array(curr_axis[ann]))
                if ann_schema['type']!='categorical':
                    ann_schema['writable']=False
                self.schema["annotations"][ax]["columns"].append(ann_schema)
        for layout in self.get_embedding_names():
            layout_schema = {"name": layout, "type": "float32", "dims": [f"{layout}_0", f"{layout}_1"]}
            self.schema["layout"]["obs"].append(layout_schema)

    def get_schema(self):
        return self.schema

    def _load_data(self, data_locator, root_embedding = None):
        # as of AnnData 0.6.19, backed mode performs initial load fast, but at the
        # cost of significantly slower access to X data.
        try:
            # there is no guarantee data_locator indicates a local file.  The AnnData
            # API will only consume local file objects.  If we get a non-local object,
            # make a copy in tmp, and delete it after we load into memory.
            with data_locator.local_handle() as lh:
                backed = "r" if self.server_config.adaptor__anndata_adaptor__backed else None

                if os.path.isdir(lh) and len(glob(lh+'/*.gz'))==0:
                    filenames = glob(lh+'/*')
                    adatas = []
                    batch = []
                    for file in filenames:
                        if os.path.isdir(file):
                            backed=False
                    
                    for file in filenames:
                        if os.path.isdir(file):
                            adata = sc.read_10x_mtx(file)
                            filt1,_ = sc.pp.filter_cells(adata,min_counts=100, inplace=False)
                            filt2,_ = sc.pp.filter_cells(adata,min_genes=100, inplace=False)
                            filt = np.logical_and(filt1,filt2)
                            adata = adata[filt].copy()
                        elif file.split('.')[-1] =='csv':
                            adata = sc.read_csv(file) 
                            adata.X = sp.sparse.csc_matrix(adata.X)
                        else:
                            adata = anndata.read_h5ad(file, backed=backed)

                        adatas.append(adata)
                        batch.append([file.split('.h5ad')[0].split('/')[-1]]*adata.shape[0])
                    adata = anndata.concat(adatas,join='inner',axis=0)
                    if "orig.ident" not in adata.obs.keys():
                        key = "orig.ident"
                    else:
                        key = f"orig.ident.{str(hex(int(time.time())))[2:]}"
                    adata.obs[key] = pd.Categorical(np.concatenate(batch))
                elif len(glob(lh+'/*.gz'))>0:
                    adata = sc.read_10x_mtx(lh)
                else:
                    adata = anndata.read_h5ad(lh, backed=backed)


                if not sparse.issparse(adata.X):
                    adata.X = sparse.csr_matrix(adata.X)

                for k in adata.layers.keys():  
                    if not sparse.issparse(adata.layers[k]):
                        adata.layers[k] = sparse.csr_matrix(adata.layers[k])

                if root_embedding is not None:
                    adata.obsm["X_root"] = adata.obsm[root_embedding]
                    if root_embedding[:2] == "X_":
                        obsp = root_embedding[2:]
                    else:
                        obsp = root_embedding
                    
                    if "N_"+obsp in adata.obsp.keys():
                        adata.obsp["N_root"] = adata.obsp["N_"+obsp]
                        adata.uns["N_root_params"] = adata.uns["N_"+obsp+"_params"]
                        del adata.obsp["N_"+obsp]
                        del adata.uns["N_"+obsp+"_params"]

                    del adata.obsm[root_embedding]
                else:
                    adata.obsm["X_root"] = np.zeros((adata.shape[0],2))
                
                adata.obs_names_make_unique()

                # cast all expressions to float32 if they're not already
                if adata.X.dtype != "float32":
                    adata.X = adata.X.astype('float32')
                for k in adata.layers.keys():
                    if adata.layers[k].dtype != "float32":
                        adata.layers[k] = adata.layers[k].astype('float32')

                self.shm_layers_csr = {}
                self.shm_layers_csc = {}
                if adata.X.getformat() == "csr":
                    self.shm_layers_csr["X"] = _create_shm_from_data(adata.X)
                    adata.X=fmt_swapper(adata.X)
                elif adata.X.getformat() != "csc":
                    adata.X=adata.X.tocsc()
                
                adata.layers["X"] = adata.X

                print("Loading and precomputing layers necessary for fast differential expression and reembedding...")
                
                # convert everything to CSC and cache all CSR into shared memory.
                shm_keys = list(self.shm_layers_csr.keys())
                for k in list(adata.layers.keys()):  
                    if k not in shm_keys:
                        if adata.layers[k].getformat() == "csr": # if csr, swap to csc and cache csr into shared memory.
                            self.shm_layers_csr[k] = _create_shm_from_data(adata.layers[k])
                            adata.layers[k] = fmt_swapper(adata.layers[k])
                        elif adata.layers[k].getformat() != "csc": # if any other format, just convert to csc.
                            adata.layers[k] = adata.layers[k].tocsc()
                                
                # cache all remaining CSC into CSR shared memory
                shm_keys = list(self.shm_layers_csr.keys())
                for key in list(adata.layers.keys()):
                    X = adata.layers[key]
                    if  key not in shm_keys: # if key not in shm_keys, then it means that it's CSC and didn't come from CSR.
                        X2 = fmt_swapper(X) # convert csc to csr
                        self.shm_layers_csr[key] = _create_shm_from_data(X2) #cache csr into shared memory

                    mean,v = sf.mean_variance_axis(X,axis=0)
                    meansq = v-mean**2
                    adata.var[f"{key};;tMean"] = mean
                    adata.var[f"{key};;tMeanSq"] = meansq
                
                for k in adata.layers.keys():
                    self.shm_layers_csc[k] = _create_shm_from_data(adata.layers[k])                    
                
                if 'orig.X' not in adata.layers.keys(): 
                    adata.layers['orig.X'] = adata.X   
                    adata.var['orig.X;;tMean'] = adata.var['X;;tMean']
                    adata.var['orig.X;;tMeanSq'] = adata.var['X;;tMeanSq']
                    self.shm_layers_csr['orig.X'] = self.shm_layers_csr['X']
                    self.shm_layers_csc['orig.X'] = self.shm_layers_csc['X']

                if adata.raw is not None:
                    X = adata.raw.X
                    mean,v = sf.mean_variance_axis(X,axis=0)
                    meansq = v-mean**2
                    adata.var[f".raw;;tMean"] = mean
                    adata.var[f".raw;;tMeanSq"] = meansq

                    del adata.raw
                    
                    if X.getformat() == "csc":
                        X1 = X
                        X2 = fmt_swapper(X)
                    elif X.getformat() == "csr":
                        X1 = fmt_swapper(X)
                        X2 = X
                    else:
                        X1 = X.tocsc()
                        X2 = X.tocsr()                        
                    adata.layers[".raw"] = X1
                    self.shm_layers_csr[".raw"] = _create_shm_from_data(X2)
                    self.shm_layers_csc[".raw"] = _create_shm_from_data(X1)

                for curr_axis in [adata.obs,adata.var]:
                    for ann in curr_axis:
                        dtype = curr_axis[ann].dtype
                        if hasattr(dtype,'numpy_dtype'):
                            dtype = dtype.numpy_dtype
                        curr_axis[ann] = curr_axis[ann].astype(dtype)

                self.data = adata

        except ValueError:
            raise DatasetAccessError(
                "File must be in the .h5ad format. Please read "
                "https://github.com/theislab/scanpy_usage/blob/master/170505_seurat/info_h5ad.md to "
                "learn more about this format. You may be able to convert your file into this format "
                "using `cellxgene prepare`, please run `cellxgene prepare --help` for more "
                "information."
            )
        except MemoryError:
            raise DatasetAccessError("Out of memory - file is too large for available memory.")
        except Exception as e:
            print(e)
            raise DatasetAccessError(
                "File not found or is inaccessible. File must be an .h5ad object. "
                "Please check your input and try again."
            )

    def _initialize_user_folders(self,userID):
        if not os.path.exists(f"{userID}/"):
            os.makedirs(f"{userID}/nnm/")
            os.makedirs(f"{userID}/emb/")
            os.makedirs(f"{userID}/params/")
            os.makedirs(f"{userID}/pca/")

            pickle.dump(self._obs_init,open(f"{userID}/obs.p",'wb'))
            for k in self._obsm_init.keys():
                k2 = "X_".join(k.split("X_")[1:])
                pickle.dump(self._obsm_init[k],open(f"{userID}/emb/{k2}.p",'wb'))
                r = self._obsp_init.get("N_"+k2,self._obsp_init.get("connectivities",None))
                p = self._uns_init.get("N_"+k2+"_params",{})
                if r is not None:
                    pickle.dump(r,open(f"{userID}/nnm/{k2}.p",'wb'))
                    pickle.dump(p,open(f"{userID}/params/{k2}.p",'wb'))
        else:
            obs = pickle.load(open(f"{userID}/obs.p",'rb'))
            for ann in obs:
                dtype = obs[ann].dtype
                if hasattr(dtype,'numpy_dtype'):
                    dtype = dtype.numpy_dtype
                obs[ann] = obs[ann].astype(dtype)     
            pickle.dump(obs,open(f"{userID}/obs.p",'wb'))


            

    def _validate_and_initialize(self):
        if anndata_version_is_pre_070():
            warnings.warn(
                "Use of anndata versions older than 0.7 will have serious issues. Please update to at "
                "least anndata 0.7 or later."
            )

        # var and obs column names must be unique
        if not self.data.obs.columns.is_unique or not self.data.var.columns.is_unique:
            raise KeyError("All annotation column names must be unique.")

        self._alias_annotation_names()
        self._validate_data_types()
        self.cell_count = self.data.shape[0]
        self.gene_count = self.data.shape[1]
        self._create_schema()

        self._obsm_init = self.data.obsm
        self._obs_init = self.data.obs
        self._uns_init = self.data.uns
        self._obsp_init = self.data.obsp

        del self.data.obs
        del self.data.obsm
        del self.data.uns
        del self.data.obsp

        self.data.obsm['X_root'] = self._obsm_init['X_root']
        self.data.obs["name_0"] = self._obs_init["name_0"]

        self._obs_init = self._obs_init.set_index("name_0")


        # heuristic
        n_values = self.data.shape[0] * self.data.shape[1]
        if (n_values > 1e8 and self.server_config.adaptor__anndata_adaptor__backed is True) or (n_values > 5e8):
            self.parameters.update({"diffexp_may_be_slow": True})


        id = (self.get_location()).encode()
        self.guest_idhash = base64.b32encode(blake2b(id, digest_size=5).digest()).decode("utf-8")
        self._initialize_user_folders(self.guest_idhash)

    def _is_valid_layout(self, arr):
        """return True if this layout data is a valid array for front-end presentation:
        * ndarray, dtype float/int/uint
        * with shape (n_obs, >= 2)
        * with all values finite or NaN (no +Inf or -Inf)
        """
        is_valid = type(arr) == np.ndarray and arr.dtype.kind in "fiu"
        is_valid = is_valid and arr.shape[0] == self.data.n_obs and arr.shape[1] >= 2
        is_valid = is_valid and not np.any(np.isinf(arr)) and not np.all(np.isnan(arr))
        return is_valid

    def _validate_data_types(self):
        # The backed API does not support interrogation of the underlying sparsity or sparse matrix type
        # Fake it by asking for a small subarray and testing it.   NOTE: if the user has ignored our
        # anndata <= 0.7 warning, opted for the --backed option, and specified a large, sparse dataset,
        # this "small" indexing request will load the entire X array. This is due to a bug in anndata<=0.7
        # which will load the entire X matrix to fullfill any slicing request if X is sparse.  See
        # user warning in _load_data().
        X0 = self.data.X[0, 0:1]
        if sparse.isspmatrix(X0) and not sparse.isspmatrix_csc(X0):
            warnings.warn(
                "Anndata data matrix is sparse, but not a CSC (columnar) matrix.  "
                "Performance may be improved by using CSC."
            )
        if self.data.X.dtype != "float32":
            warnings.warn(
                f"Anndata data matrix is in {self.data.X.dtype} format not float32. " f"Precision may be truncated."
            )
        for ax in Axis:
            curr_axis = getattr(self.data, str(ax))
            for ann in curr_axis:
                datatype = curr_axis[ann].dtype
                downcast_map = {
                    "int64": "int32",
                    "uint32": "int32",
                    "uint64": "int32",
                    "float64": "float32",
                }
                if datatype in downcast_map:
                    warnings.warn(
                        f"Anndata annotation {ax}:{ann} is in unsupported format: {datatype}. "
                        f"Data will be downcast to {downcast_map[datatype]}."
                    )
                if isinstance(datatype, CategoricalDtype):
                    category_num = len(curr_axis[ann].dtype.categories)
                    if category_num > 500 and category_num > self.dataset_config.presentation__max_categories:
                        warnings.warn(
                            f"{str(ax).title()} annotation '{ann}' has {category_num} categories, this may be "
                            f"cumbersome or slow to display. We recommend setting the "
                            f"--max-category-items option to 500, this will hide categorical "
                            f"annotations with more than 500 categories in the UI"
                        )

    def annotation_to_fbs_matrix(self, axis, fields=None, labels=None):
        if axis == Axis.OBS:
            if labels is not None and not labels.empty:
                labels["name_0"] = self.data.obs["name_0"]
                df = labels
            else:
                df = self.data.obs
        else:
            df = self.data.var

        if fields is not None and len(fields) > 0:
            df = df[fields]
        return encode_matrix_fbs(df, col_idx=df.columns)

    def get_embedding_names(self):
        """
        Return pre-computed embeddings.

        function:
            a) generate list of default layouts
            b) validate layouts are legal.  remove/warn on any that are not
            c) cap total list of layouts at global const MAX_LAYOUTS
        """
        # load default layouts from the data.
        layouts = self.dataset_config.embeddings__names

        if layouts is None or len(layouts) == 0:
            layouts = [key[2:] for key in self.data.obsm_keys() if type(key) == str and key.startswith("X_")]

        # remove invalid layouts
        valid_layouts = []
        obsm_keys = self.data.obsm_keys()
        for layout in layouts:
            layout_name = f"X_{layout}"
            if layout_name not in obsm_keys:
                warnings.warn(f"Ignoring unknown layout name: {layout}.")
            elif not self._is_valid_layout(self.data.obsm[layout_name]):
                warnings.warn(f"Ignoring layout due to malformed shape or data type: {layout}")
            else:
                valid_layouts.append(layout)
        # cap layouts to MAX_LAYOUTS
        return valid_layouts[0:MAX_LAYOUTS]

    def get_embedding_array(self, ename, dims=2):
        annotations = self.dataset_config.user_annotations        
        userID = f"{annotations._get_userdata_idhash(self)}"
        try:
            full_embedding = pickle.load(open(f"{userID}/emb/{ename}.p",'rb'))
        except:
            full_embedding = self._obsm_init[f"X_{ename}"]
        return full_embedding[:, 0:dims]

    def get_colors(self):
        return convert_anndata_category_colors_to_cxg_category_colors(self.data)

    def get_X_array(self, col_idx, layer="X", logscale=False):
        def bisym_log_transform(x):
             return np.sign(x)*np.log(1+np.abs(x))

        #if row_idx is None:
        #    row_idx = np.arange(self.data.shape[0])
        if layer == "X":
            XI = self.data.X
        else:
            XI = self.data.layers[layer]

        if col_idx is None:
            col_idx = np.arange(self.data.shape[1])        
        
        if col_idx.size == 1:
            i1 = col_idx[0]
                            
            d = XI.data[XI.indptr[i1] : XI.indptr[i1 + 1]]
            i = XI.indices[XI.indptr[i1] : XI.indptr[i1 + 1]]
            x = np.zeros(XI.shape[0])
            x[i] = d
            x=x[:,None]
            if logscale:
                x = bisym_log_transform(x)
            #x=x[row_idx][:,None]                
        else:
            x = XI[:,col_idx]
            if logscale:
                if sparse.issparse(x):
                    x.data[:] = bisym_log_transform(x.data)
                else:
                    x = bisym_log_transform(x)
        return x

    def get_shape(self):
        return self.data.shape

    def query_var_array(self, term_name):
        return getattr(self.data.var, term_name)

    def query_obs_array(self, term_name):
        return getattr(self.data.obs, term_name)

    def get_obs_index(self):
        name = self.server_config.single_dataset__obs_names
        if name is None:
            return self.original_obs_index
        else:
            return self.data.obs[name]

    def get_obs_columns(self):
        return self.data.obs.columns

    def get_obs_keys(self):
        # return list of keys
        return self.data.obs.keys().to_list()

    def get_var_keys(self):
        # return list of keys
        return self.data.var.keys().to_list()




import yaml

default_config = """
server:
  app:
    verbose: false
    debug: false
    host: localhost
    port : null
    open_browser: false
    force_https: false
    flask_secret_key: null
    generate_cache_control_headers: false

  authentication:
    # The authentication types may be "none" or "session"
    # none:  No authentication support, features like user_annotations must not be enabled.
    # session:  A session based userid is automatically generated. (no params needed)
    type: session
    insecure_test_environment: false

  single_dataset:
    # If datapath is set, then cellxgene with serve a single dataset located at datapath.
    datapath: null
    obs_names: null
    var_names: null
    about: null
    title: null

  data_locator:
    s3:
      # s3 region name.
      #   if true, then the s3 location is automatically determined from the datapath or dataroot.
      #   if false/null, then do not set.
      #   if a string, then use that value (e.g. us-east-1).
      region_name: true

  adaptor:
    anndata_adaptor:
      backed: false

  limits:
    column_request_max: 32
    diffexp_cellcount_max: null


dataset:
  app:
    # Scripts can be a list of either file names (string) or dicts containing keys src, integrity and crossorigin.
    # these will be injected into the index template as script tags with these attributes set.
    scripts: []
    # Inline scripts are a list of file names, where the contents of the file will be injected into the index.
    inline_scripts: []

    # allow authentication support
    authentication_enable: true

  presentation:
    max_categories: 1000
    custom_colors: true

  user_annotations:
    enable: true
    type: local_file_csv
    local_file_csv:
      directory: null
      file: null            # annotations file name
      gene_sets_file: null  # gene sets file name
      reembed_parameters_file: null # reembed parameters file name
    ontology:
      enable: false
      obo_location: null
    gene_sets:
      readonly: false       # gene sets CRUD enabled/disabled

  embeddings:
    names : []

  diffexp:
    enable: true
    lfc_cutoff: 0.01
    top_n: 10

external:
  # You can retrieve configuration parameters from this config file, the environment,
  # the AWS secrets manager, or from the "cellxgene launch" command line arguments.
  # They are applied in that order, meaning that if a parameter is defined in more
  # than one location, the last one applied takes effect.

  # environment variables:
  # This section describes how to map environment variables to configuration parameters.
  # The format is a list defining an environment variable.
  # Each entry in the list is a dictionary with three entries:
  # name:  the name of the environment variable
  # path: the path within the cellxgene configuration to update.
  # required: (default=False) a boolean.  If true, then it is an error if the environment variable is not set.

  environment:
     - name: CXG_SECRET_KEY
       path: [server, app, flask_secret_key]
       required: false

  # AWS Secrets Manager
  # This section describes how to map aws secrets to configuration parameters.
  # The format is the region for the secrets manager, then a list of secrets.
  # each secret has a name, and a list of values.
  # Each entry in the list of values is a dictionary with three entries:
  # key:  the key of the aws secret.
  # path: the path within the cellxgene configuration to update.
  # required: (default=False) a boolean.  If true, then it is an error if the key does not exist in the secret.
  #
  # example:
  # aws_secrets_manager:
  #   region: us-west-2
  #    - name: my_first_secret
  #      values:
  #      - key: flask_secret_key
  #        path: [server, app, flask_secret_key]
  #        required: true
  #      - key: db_uri
  #        path: [dataset, user_annotations, db_uri]
  #        required: true
  #    - name:  my_auth_secret
  #      values:
  #      - key: client_secret
  #        path: [server, authentication, client_secret]
  #        required: true
  #      - key: client_id
  #        path: [server, authentication, client_id]
  #        required: true

  aws_secrets_manager:
    region: null
    secrets: []
"""


def get_default_config():
    return yaml.load(default_config, Loader=yaml.Loader)



from http import HTTPStatus
from flask import make_response, jsonify

from backend.server import __version__ as cellxgene_version
from backend.common.utils.data_locator import DataLocator


def _is_accessible(path, config):
    if path is None:
        return True

    try:
        dl = DataLocator(path, region_name=config.data_locator__s3__region_name)
        return dl.exists()
    except RuntimeError:
        return False


def health_check(config):
    """
    simple health check - return HTTP response.
    See https://tools.ietf.org/id/draft-inadarei-api-health-check-01.html
    """
    health = {"status": None, "version": "1", "releaseID": cellxgene_version}

    server_config = config.server_config
    check = _is_accessible(server_config.single_dataset__datapath, server_config)

    health["status"] = "pass" if check else "fail"
    code = HTTPStatus.OK if health["status"] == "pass" else HTTPStatus.BAD_REQUEST
    response = make_response(jsonify(health), code)
    response.headers["Content-Type"] = "application/health+json"
    return response



import json
import re
import os
import sys

import pandas as pd
import yaml


def _is_null(v):
    """Return True if v is null, for one of the multiple ways a "null" value shows up in an h5ad."""
    return pd.isnull(v) or (hasattr(v, "__len__") and len(v) == 0)


def _validate_stringified_list_of_dicts(s):
    """Verify that a string can be parsed into a list.

    We have some types that are lists of dicts. Those cannot be stored directly in an h5ad, so we have to
    json.dumps them. This verifies that we can load them back.
    """

    try:
        list_ = json.loads(s)
        if not isinstance(list_, list):
            return False
        for el in list_:
            if not isinstance(el, dict):
                return False
        return True
    except (json.JSONDecodeError, TypeError):
        pass
    return False


def _validate_human_readable_string(s):
    """Verify that a string is human-readable.

    There are parts of the schema where a "human-readable" string is required. "Human-readable" is kind
    of vague and subjective. I feel like I can read many strings. So here we just check for the main ways
    that fails: someone puts in an ontology term id or and ensembl gene/transcript id.

    Returns False if s is not a string or is one of those bad string types.
    """

    return isinstance(s, str) and (not re.match(r"[A-Z]\w+:\d+", s)) and (not re.match(r"ENS[GT]\d+$", s))


def _validate_curie(c, prefixes):
    """Verify that a string is a valid compact URI, like EFO:000001. If prefixes is not empty, make sure the
    prefix of the curies is in prefixes.
    """

    if not c:
        return True

    match = re.match(r"([A-Z]\w+):\d+$", c)

    if prefixes:
        return match and match.group(1) in prefixes
    else:
        return match


def _validate_suffixed_curie(c, prefixes):
    """Verify that a string is a compact URI with an optional suffix like 'EFO:00001 (cell culture)'"""

    # Pull off the suffix
    suffix = re.findall(r"\ \(.*\)$", c)
    if suffix:
        c = c[: -len(suffix[0])]
    return _validate_curie(c, prefixes)


def _validate_column(column, column_name, df_name, schema_def):
    """Given a schema definition and the column of a dataframe, verify that the column satifies
    the schema.
    """

    errors = []

    if schema_def.get("unique"):
        if column.nunique() != len(column):
            errors.append(f"Column {column_name} in dataframe {df_name} is not unique.")

    if "nullable" in schema_def and not schema_def["nullable"]:
        if any(_is_null(v) for v in column):
            errors.append(f"Column {column_name} in dataframe {df_name} contains empty values.")

    if schema_def.get("type") == "human-readable string":
        non_readables = [v for v in column if not _validate_human_readable_string(v)]
        if non_readables:
            errors.append(
                f"Column {column_name} in dataframe {df_name} contains non-human-readable "
                f"values like {non_readables[0]}"
            )

    if schema_def.get("type") in ("curie", "suffixed curie"):
        validation_func = _validate_curie if schema_def.get("type") == "curie" else _validate_suffixed_curie
        non_valid_curies = [v for v in column if not validation_func(v, schema_def.get("prefixes"))]
        if non_valid_curies:
            errors.append(
                f"Column {column_name} in dataframe {df_name} contains invalid ontology values like "
                f"{non_valid_curies[0]}."
            )
            if "prefixes" in schema_def:
                errors[-1] += f" Values must be curies from one of these ontologies {schema_def['prefixes']}."

    if "enum" in schema_def:
        bad_enums = [v for v in column if v not in schema_def["enum"]]
        if bad_enums:
            errors.append(
                f"Column {column_name} in dataframe {df_name} contains unpermitted values like "
                f"{bad_enums[0]}. Values must be one of {schema_def['enum']}."
            )

    return errors


def _validate_dict(dict_, dict_name, schema_def):
    """Given a schema definition and dict, verify that the dict satifies the schema."""

    errors = []

    for key in schema_def.get("keys", []):
        if key not in dict_:
            errors.append(f"{dict_name} is missing key {key}.")
        elif schema_def["keys"][key]:
            if schema_def["keys"][key]["type"] == "stringified list of dicts":
                if not _validate_stringified_list_of_dicts(dict_[key]):
                    errors.append(
                        f"Key {key} in {dict_name} should be a JSON-encoded list of dicts, but it is {dict_[key]}"
                    )
            elif schema_def["keys"][key]["type"] == "dict":
                errors.extend(_validate_dict(dict_[key], key, schema_def["keys"][key]))
            elif schema_def["keys"][key]["type"] == "curie":
                if not _validate_curie(dict_[key], schema_def["keys"][key]["prefixes"]):
                    errors.append(f"Key {key} in {dict_name} contains invalid ontology value.")
            if "nullable" in schema_def["keys"][key] and not schema_def["keys"][key]["nullable"]:
                if _is_null(dict_[key]):
                    errors.append(f"Key {key} in dict {dict_name} is an empty value.")

    return errors


def _validate_dataframe(df, df_name, schema_def):
    """Given a dataframe and schema definition, verify that the dataframe follows the schema."""

    errors = []

    if "index" in schema_def:
        errors.extend(_validate_column(df.index, "index", df_name, schema_def["index"]))

    for column in schema_def.get("columns", []):
        if column not in df.columns:
            errors.append(f"Dataframe {df_name} is missing column {column}.")
        else:
            errors.extend(_validate_column(df[column], column, df_name, schema_def["columns"][column]))

    return errors


def get_schema_definition(version):
    """Look up and read a schema definition based on a version number like "1.0.0"."""

    path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "schema_definitions", version.replace(".", "_") + ".yaml"
    )

    if not os.path.isfile(path):
        raise ValueError(f"No definition for version {version} found.")

    return yaml.load(open(path), Loader=yaml.FullLoader)


def deep_check(adata, schema_def):
    """Perform a "deep" check of the AnnData object using the schema definition.

    This checks all the columns and unstructured metadata rather than just the version.

    Returns a list of error messages. If that list is empty, the object passed validation.
    """

    errors = []

    for component, component_def in schema_def["components"].items():
        if component_def["type"] == "dataframe":
            errors.extend(_validate_dataframe(getattr(adata, component), component, component_def))
        elif component_def["type"] == "dict":
            errors.extend(_validate_dict(getattr(adata, component), component, component_def))
        else:
            raise ValueError(f"Unexpected component type {component['type']}")

    return errors


def validate_adata(adata, shallow):
    """Validate an AnnData object. If shallow, just check that the required version information is
    present.
    """

    # Does it have the version information written into uns?
    if "version" not in adata.uns_keys() or "corpora_schema_version" not in adata.uns["version"]:
        print("AnnData file is missing corpora version information")
        return False

    # We can stop here if it's a "shallow" check, that is, if we're just
    # checking that version is present.
    if shallow:
        return True

    schema_def = get_schema_definition(adata.uns["version"]["corpora_schema_version"])

    errors = deep_check(adata, schema_def)

    for error in errors:
        print(error)

    return not errors


def validate(h5ad_path, shallow=False):
    """Entry point for validation."""

    try:
        import scanpy
    except ImportError:
        raise ImportError("scanpy must be installed for cellxgene schema")

    try:
        adata = scanpy.read_h5ad(h5ad_path, backed="r")
    except (OSError, TypeError):
        print(f"Unable to open {h5ad_path} with scanpy.")
        sys.exit(1)

    if not validate_adata(adata, shallow):
        sys.exit(1)



import datetime
import logging
from functools import wraps
from http import HTTPStatus

from flask import (
    Flask,
    current_app,
    make_response,
    jsonify,
    render_template,
    Blueprint,
    request,
    send_from_directory,
    session,
    after_this_request,
    send_file
)
from flask_restful import Api, Resource
import backend.server.common.rest as common_rest
from backend.common.errors import DatasetAccessError, RequestException
from backend.server.common.health import health_check
from backend.common.utils.utils import Float32JSONEncoder
import os

webbp = Blueprint("webapp", "backend.server.common.web", template_folder="templates")

ONE_WEEK = 7 * 24 * 60 * 60

   
def auth0_token_required(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        token = 'profile' in session
        # return 401 if token is not passed
        if not token and current_app.hosted_mode:
            return jsonify({'message' : 'Authorization missing.'}), 401
  
        return  f(*args, **kwargs)
  
    return decorated

def desktop_mode_only(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if current_app.hosted_mode:
            return jsonify({'message' : 'Feature only available in desktop mode.'}), 401
  
        return  f(*args, **kwargs)
  
    return decorated    

def hosted_mode_only(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if not current_app.hosted_mode:
            return jsonify({'message' : 'Feature only available in hosted mode.'}), 401
  
        return  f(*args, **kwargs)
  
    return decorated    

def _cache_control(always, **cache_kwargs):
    """
    Used to easily manage cache control headers on responses.
    See Werkzeug for attributes that can be set, eg, no_cache, private, max_age, etc.
    https://werkzeug.palletsprojects.com/en/1.0.x/datastructures/#werkzeug.datastructures.ResponseCacheControl
    """

    def inner_cache_control(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            response = make_response(f(*args, **kwargs))
            if not always and not current_app.app_config.server_config.app__generate_cache_control_headers:
                return response
            if response.status_code >= 400:
                return response
            for k, v in cache_kwargs.items():
                setattr(response.cache_control, k, v)
            return response

        return wrapper

    return inner_cache_control


def cache_control(**cache_kwargs):
    """ config driven """
    return _cache_control(False, **cache_kwargs)


def cache_control_always(**cache_kwargs):
    """ always generate headers, regardless of the config """
    return _cache_control(True, **cache_kwargs)


@webbp.route("/", methods=["GET"])
def dataset_index():
    app_config = current_app.app_config

    dataset_config = app_config.get_dataset_config()
    scripts = dataset_config.app__scripts
    inline_scripts = dataset_config.app__inline_scripts

    try:
        args = {"SCRIPTS": scripts, "INLINE_SCRIPTS": inline_scripts}
        return render_template("index.html", **args)

    except DatasetAccessError as e:
        return common_rest.abort_and_log(
            e.status_code, f"Invalid dataset: {e.message}", loglevel=logging.INFO, include_exc_info=True
        )


@webbp.errorhandler(RequestException)
def handle_request_exception(error):
    return common_rest.abort_and_log(error.status_code, error.message, loglevel=logging.INFO, include_exc_info=True)


def requires_authentication(func):
    @wraps(func)
    def wrapped_function(self, *args, **kwargs):
        auth = current_app.auth
        if auth.is_user_authenticated():
            return func(self, *args, **kwargs)
        else:
            return make_response("not authenticated", HTTPStatus.UNAUTHORIZED)

    return wrapped_function


def rest_get_data_adaptor(func):
    @wraps(func)
    def wrapped_function(self):
        try:
            return func(self, current_app.data_adaptor)
        except DatasetAccessError as e:
            return common_rest.abort_and_log(
                e.status_code, f"Invalid dataset: {e.message}", loglevel=logging.INFO, include_exc_info=True
            )

    return wrapped_function


class HealthAPI(Resource):
    @cache_control_always(no_store=True)
    def get(self):
        config = current_app.app_config
        return health_check(config)


class SchemaAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.schema_get(data_adaptor)

class InitializeUserAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.initialize_user(data_adaptor)

class ConfigAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.config_get(current_app.app_config, data_adaptor)


class UserInfoAPI(Resource):
    @cache_control_always(no_store=True)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.userinfo_get(current_app.app_config, data_adaptor)


class AnnotationsObsAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.annotations_obs_get(request, data_adaptor)

    @requires_authentication
    @cache_control(no_store=True)
    @auth0_token_required
    @rest_get_data_adaptor
    def put(self, data_adaptor):
        return common_rest.annotations_obs_put(request, data_adaptor)

class UserInfoAuth0API(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    def get(self):
        if 'profile' in session:
            return make_response(jsonify({"response": session['profile']}), HTTPStatus.OK)
        else:
            return make_response(jsonify({"response": None}), HTTPStatus.OK)            

class HostedModeAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    def get(self):
        return make_response(jsonify({"response": current_app.hosted_mode}), HTTPStatus.OK)

class SendFileAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    @auth0_token_required
    def get(self,data_adaptor):
        field = request.args.get("path", None)
    
        annotations = data_adaptor.dataset_config.user_annotations        
        userID = f"{annotations._get_userdata_idhash(data_adaptor)}"             
        assert (userID in field)
        @after_this_request
        def f(response):
            try:
                os.remove(field)
            except Exception as error:
                print(error)
            return response    
        
        return send_file(field,as_attachment=True)    

class AnnotationsVarAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.annotations_var_get(request, data_adaptor)


class DataVarAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def put(self, data_adaptor):
        return common_rest.data_var_put(request, data_adaptor)

    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.data_var_get(request, data_adaptor)


class ColorsAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.colors_get(data_adaptor)


class DiffExpObsAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def post(self, data_adaptor):
        return common_rest.diffexp_obs_post(request, data_adaptor)

class PreprocessAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    @desktop_mode_only
    def put(self, data_adaptor):
        return common_rest.preprocess_put(request, data_adaptor)


class LayoutObsAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.layout_obs_get(request, data_adaptor)

class SankeyPlotAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    def put(self, data_adaptor):
        return common_rest.sankey_data_put(request, data_adaptor)

class OutputAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    def put(self, data_adaptor):
        return common_rest.output_data_put(request, data_adaptor)

class DownloadAnndataAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def put(self, data_adaptor):
        return common_rest.save_data_put(request, data_adaptor)

class DownloadMetadataAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    def put(self, data_adaptor):
        return common_rest.save_metadata_put(request, data_adaptor)

class DownloadGenedataAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    def put(self, data_adaptor):
        return common_rest.save_genedata_put(request, data_adaptor)

class DeleteObsmAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def put(self, data_adaptor):
        return common_rest.delete_obsm_put(request, data_adaptor)

class RenameObsmAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def put(self, data_adaptor):
        return common_rest.rename_obsm_put(request, data_adaptor)

class LeidenClusterAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def put(self, data_adaptor):
        return common_rest.leiden_put(request, data_adaptor)

class ReembedParametersObsmAPI(Resource):
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    def put(self, data_adaptor):
        return common_rest.reembed_parameters_obsm_put(request, data_adaptor)

class GeneInfoAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.gene_info_get(request, data_adaptor)

class ReembedParametersAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.reembed_parameters_get(request, data_adaptor)

    @requires_authentication
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def put(self, data_adaptor):
        return common_rest.reembed_parameters_put(request, data_adaptor)

class GenesetsAPI(Resource):
    @cache_control(public=True, max_age=ONE_WEEK)
    @rest_get_data_adaptor
    def get(self, data_adaptor):
        return common_rest.genesets_get(request, data_adaptor)

    @requires_authentication
    @cache_control(no_store=True)
    @rest_get_data_adaptor
    @auth0_token_required
    def put(self, data_adaptor):
        return common_rest.genesets_put(request, data_adaptor)

class SummarizeVarAPI(Resource):
    @rest_get_data_adaptor
    @cache_control(public=True, max_age=ONE_WEEK)
    def get(self, data_adaptor):
        return common_rest.summarize_var_get(request, data_adaptor)

    @rest_get_data_adaptor
    @cache_control(no_store=True)
    @auth0_token_required
    def post(self, data_adaptor):
        return common_rest.summarize_var_post(request, data_adaptor)


def get_api_base_resources(bp_base):
    """Add resources that are accessed from the api url"""
    api = Api(bp_base)

    # Diagnostics routes
    api.add_resource(HealthAPI, "/health")
    return api


def get_api_dataroot_resources(bp_dataroot):
    """Add resources that refer to a dataset"""
    api = Api(bp_dataroot)

    def add_resource(resource, url):
        """convenience function to make the outer function less verbose"""
        api.add_resource(resource, url)

    # Initialization routes
    add_resource(SchemaAPI, "/schema")
    add_resource(InitializeUserAPI, "/initialize")
    add_resource(SendFileAPI, "/sendFile")
    add_resource(ConfigAPI, "/config")
    add_resource(UserInfoAPI, "/userinfo")
    add_resource(UserInfoAuth0API, "/userInfo")
    add_resource(HostedModeAPI, "/hostedMode")
    add_resource(GeneInfoAPI, "/geneInfo")
    # Data routes
    add_resource(AnnotationsObsAPI, "/annotations/obs")
    add_resource(AnnotationsVarAPI, "/annotations/var")
    add_resource(DataVarAPI, "/data/var")
    add_resource(GenesetsAPI, "/genesets")
    
    add_resource(ReembedParametersAPI, "/reembed-parameters")
    add_resource(ReembedParametersObsmAPI, "/reembed-parameters-obsm")

    #add_resource(SankeyPlotAPI, "/sankey")
    #add_resource(OutputAPI, "/output")
    #add_resource(DownloadAnndataAPI, "/downloadAnndata")
    add_resource(DownloadMetadataAPI, "/downloadMetadata")
    add_resource(DownloadGenedataAPI, "/downloadGenedata")
    #add_resource(LeidenClusterAPI, "/leiden")
    add_resource(DeleteObsmAPI, "/layout/obsm")
    add_resource(RenameObsmAPI, "/layout/rename")
    #add_resource(PreprocessAPI, "/preprocess")
    add_resource(SummarizeVarAPI, "/summarize/var")
    # Display routes
    add_resource(ColorsAPI, "/colors")
    # Computation routes
    #add_resource(DiffExpObsAPI, "/diffexp/obs")
    add_resource(LayoutObsAPI, "/layout/obs")
    return api


class Server:
    @staticmethod
    def _before_adding_routes(app, app_config):
        """ will be called before routes are added, during __init__.  Subclass protocol """
        pass

    def __init__(self, app_config):
        self.app = Flask(__name__, static_folder=None)
        self._before_adding_routes(self.app, app_config)
        self.app.json_encoder = Float32JSONEncoder
        server_config = app_config.server_config

        # enable session data
        self.app.permanent_session_lifetime = datetime.timedelta(days=50 * 365)

        # Config
        secret_key = server_config.app__flask_secret_key
        self.app.config.update(SECRET_KEY=secret_key)

        self.app.register_blueprint(webbp)

        api_version = "/api/v0.2"
        api_path = "/"

        bp_base = Blueprint("bp_base", __name__, url_prefix=api_path)
        base_resources = get_api_base_resources(bp_base)
        self.app.register_blueprint(base_resources.blueprint)

        bp_api = Blueprint("api", __name__, url_prefix=f"{api_path}{api_version}")
        resources = get_api_dataroot_resources(bp_api)
        self.app.register_blueprint(resources.blueprint)
        self.app.add_url_rule(
            "/static/<path:filename>",
            "static_assets",
            view_func=lambda filename: send_from_directory("../common/web/static", filename),
            methods=["GET"],
        )

        self.app.data_adaptor = server_config.data_adaptor
        self.app.app_config = app_config

        auth = server_config.auth
        self.app.auth = auth
        if auth.requires_client_login():
            auth.add_url_rules(self.app)
        auth.complete_setup(self.app)



"""Methods for working with ontologies and the OLS."""
from urllib.parse import quote_plus

import requests

OLS_API_ROOT = "http://www.ebi.ac.uk/ols/api"

# Curie means something like CL:0000001


def _ontology_name(curie):
    """Get the name of the ontology from the curie, CL or UBERON for example."""
    return curie.split(":")[0]


def _ontology_value(curie):
    """Get the id component of the curie, 0000001 from CL:0000001 for example."""
    return curie.split(":")[1]


def _double_encode(url):
    """Double url encode a url. This is required by the OLS API."""
    return quote_plus(quote_plus(url))


def _iri(curie):
    """Get the iri from a curie. This is a bit hopeful that they all map to purl.obolibrary.org"""
    if _ontology_name(curie) == "EFO":
        return f"http://www.ebi.ac.uk/efo/EFO_{_ontology_value(curie)}"
    return f"http://purl.obolibrary.org/obo/{_ontology_name(curie)}_{_ontology_value(curie)}"


class OntologyLookupError(Exception):
    """Exception for some problem with looking up ontology information."""


def _ontology_info_url(curie):
    """Get the to make a GET to to get information about an ontology term."""

    # If the curie is empty, just return an empty string. This happens when there is no
    # valid ontology value.
    if not curie:
        return ""
    else:
        return f"{OLS_API_ROOT}/ontologies/{_ontology_name(curie)}/terms/{_double_encode(_iri(curie))}"


def get_ontology_label(curie):
    """For a given curie like 'CL:1000413', get the label like 'endothelial cell of artery'"""

    url = _ontology_info_url(curie)

    if not url:
        return ""

    response = requests.get(url)

    if not response.ok:
        raise OntologyLookupError(
            f"Curie {curie} lookup failed, got status code {response.status_code}: {response.text}"
        )
    return response.json()["label"]


def lookup_candidate_term(label, ontology="cl", method="select"):
    """Lookup candidate terms for a label. This is useful when there is an existing label in a
    submitted dataset, and you want to find an appropriate ontology term.

    Args:
      label: the label to find ontology terms for
      ontology: the ontology to search in, cl or uberon or efo for example
      method: select or search. search provides much broader results

    Returns:
      list of (curie, label) tuples returned by OLS
    """
    # using OLS REST API [https://www.ebi.ac.uk/ols/docs/api]
    url = f"{OLS_API_ROOT}/{method}?q={quote_plus(label)}&ontology={ontology.lower()}"
    response = requests.get(url)

    if not response.ok:
        raise OntologyLookupError(
            f"Label {label} lookup failed, got status code {response.status_code}: {response.text}"
        )

    return [(r["obo_id"], r["label"]) for r in response.json()["response"]["docs"]]
